# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: timeseries_field_corrections
   :platform: Unix
   :synopsis: A Plugin to apply a simple dark and flatfield correction to some
       raw timeseries data

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

:synopsis: A Plugin to apply a simple dark and flatfield correction to some
       raw timeseries data

"""

from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.corrections.base_correction import BaseCorrection

import numpy as np

from savu.plugins.utils import register_plugin


class TimeseriesFieldCorrections(BaseCorrection, CpuPlugin):
    """
    A Plugin to apply a simple dark and flatfield correction to some
    raw timeseries data
    """

    def __init__(self):
        super(TimeseriesFieldCorrections,
              self).__init__("TimeseriesFieldCorrections")
        # TODO these should probably be parameters
        self.LOW_CROP_LEVEL = 0.0
        self.HIGH_CROP_LEVEL = 2.0
        self.WARN_PROPORTION = 0.05  # 5%
        self.flag_low_warning = False
        self.flag_high_warning = False

    def pre_process(self):
        inData = self.get_in_datasets()[0]
        self.dark = inData.data.dark_mean()
        self.flat = inData.data.flat_mean()
        self.flat_minus_dark = self.flat - self.dark

        self.flat_minus_dark = self.flat - self.dark
        det_dims = [inData.get_data_dimension_by_axis_label('detector_y'),
                    inData.get_data_dimension_by_axis_label('detector_x')]

        self.convert_size = \
            lambda x, sl: x[[sl[d] for d in det_dims]]

    def process_frames(self, data):
        data = data[0]
        dark = self.convert_size(self.dark, self.get_current_slice_list()[0])
        flat_minus_dark = self.convert_size(
            self.flat_minus_dark, self.get_current_slice_list()[0])
        data = np.nan_to_num((data-dark)/flat_minus_dark)

        # make high and low crop masks
        low_crop = data < self.LOW_CROP_LEVEL
        high_crop = data > self.HIGH_CROP_LEVEL

        # flag if those masks include a large proportion of pixels, as this
        # may indicate a failure
        if ((float(low_crop.sum()) / low_crop.size) > self.WARN_PROPORTION):
            self.flag_low_warning = True
        if ((float(high_crop.sum()) / high_crop.size) > self.WARN_PROPORTION):
            self.flag_high_warning = True

        # Set all cropped values to the crop level
        data[low_crop] = self.LOW_CROP_LEVEL
        data[high_crop] = self.HIGH_CROP_LEVEL

        return data

    def executive_summary(self):
        summary = []
        if self.flag_high_warning:
            summary.append(("WARNING: over %i%% of pixels are being clipped as " +
                           "they have %f times the intensity of the provided flat field. "+
                           "Check your Dark and Flat correction images")
                           % (self.WARN_PROPORTION*100, self.HIGH_CROP_LEVEL))

        if self.flag_low_warning:
            summary.append(("WARNING: over %i%% of pixels are being clipped as " +
                           "they below the expected lower corrected threshold of  %f. " +
                           "Check your Dark and Flat correction images")
                           % (self.WARN_PROPORTION*100, self.LOW_CROP_LEVEL))

        if len(summary) > 0:
            return summary

        return ["Nothing to Report"]



# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: region_grow3D
   :platform: Unix
   :synopsis: Fast region-grow type segmentation in 3D from Larix toolbox to evolve a given mask

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.multi_threaded_plugin import MultiThreadedPlugin
from savu.plugins.utils import register_plugin

from larix.methods.segmentation import REGION_GROW
import numpy as np

@register_plugin
class RegionGrow3d(Plugin, MultiThreadedPlugin):
    """
    Fast 3D segmentation by evolving the user-given mask, the initialised mask should be set \
    in the central part of the object to be segmented.

    :param threshold: parameter to control mask propagation. Default: 1.0.
    :param method: a method to collect statistics from the given mask (mean, median, value). Default: 'mean'.
    :param iterations: number of iterations. Default: 500.
    :param connectivity: the connectivity of the local neighbourhood, choose 4, 6, 8 or 26. Default: 6.
    :param out_datasets: The default names . Default: ['MASK_RG_EVOLVED'].
    """

    def __init__(self):
        super(RegionGrow3d, self).__init__("RegionGrow3d")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        
        getall = ['VOLUME_XZ', 'voxel_y']
        in_pData[0].plugin_data_setup('VOLUME_3D', 'single', getall=getall)
        in_pData[1].plugin_data_setup('VOLUME_3D', 'single', getall=getall) # the initialised mask

        out_dataset[0].create_dataset(in_dataset[0], dtype=np.uint8)
        out_pData[0].plugin_data_setup('VOLUME_3D', 'single', getall=getall)

    def pre_process(self):
        # extract given parameters
        self.threshold = self.parameters['threshold']
        self.iterations = self.parameters['iterations']
        self.connectivity = self.parameters['connectivity']
        if (str(self.parameters['method']) == 'mean'):
            self.method = 'mean'
        elif (str(self.parameters['method']) == 'median'):
            self.method = 'median'
        else:
            self.method = 'value'

    def process_frames(self, data):
        input_temp = data[0]
        indices = np.where(np.isnan(input_temp))
        input_temp[indices] = 0.0

        pars = {'input_data' : input_temp, # input grayscale image
        'maskData' : np.uint8(data[1]),    # generated initialisation mask
        'threhsold' : self.threshold,      # threhsold controls where evolution stops (>=1)
        'iterationsNumb' : self.iterations,# the number of iterations (depends on the size of the phase)
        'connectivity' : self.connectivity,# voxel connectivity rule, choose between 4 (2D), 6, 8 (2D), and 26
        'method' : self.method}            # method to collect statistics from the mask (mean. median, value)

        # run the method
        if (np.sum(data[1]) > 0):
            mask_evolve = REGION_GROW(pars['input_data'], pars['maskData'],\
                           pars['threhsold'], pars['iterationsNumb'],\
                           pars['connectivity'], pars['method'])
        else:
            mask_evolve = np.uint8(np.zeros(np.shape(data[0])))
        return mask_evolve

    def nInput_datasets(self):
        return 2
    def nOutput_datasets(self):
        return 1



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: i13fluo_loader
   :platform: Unix
   :synopsis: A class for loading xrf data

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.loaders.base_loader import BaseLoader
import numpy as np
from savu.plugins.utils import register_plugin
import logging
import h5py


@register_plugin
class I13FluoLoader(BaseLoader):
    """
    A class to load tomography data from an NXstxm file
    :param fluo_offset: fluo scale offset. Default: 0.0.
    :param fluo_gain: fluo gain. Default: 0.01.
    :param is_tomo: The mono energy. Default: True.
    :param theta_step: The theta step. Default:1.0.
    :param theta_start: The theta start. Default: -90.0.
    :param theta_end: The theta end. Default: 90.0.
    :param mono_energy: The mono energy. Default: 11.8.
    """

    def __init__(self, name='I13FluoLoader'):
        super(I13FluoLoader, self).__init__(name)

    def setup(self):
        """
         Define the input nexus file

        :param path: The full path of the NeXus file to load.
        :type path: str
        """

        data_str = 'entry1/instrument/xmapMca/fullSpectrum'
        exp = self.exp
        data_obj = exp.create_data_object("in_data", 'fluo')
        data_obj.backing_file = \
            h5py.File(exp.meta_data.get("data_file"), 'r')
        data_obj.data = data_obj.backing_file[data_str]
        data_obj.set_shape(data_obj.data.shape)
        npts = data_obj.data.shape[-1]
        gain = self.parameters["fluo_gain"]
        energy = np.arange(self.parameters["fluo_offset"], gain*npts, gain)
        data_obj.meta_data.set("energy", energy)
        labels = []
        ### set the x
        rotation_angle = None
        if self.parameters['is_tomo']:
            rotation_angle = np.arange(self.parameters['theta_start'],self.parameters['theta_end'], self.parameters['theta_step'])
            labels.append('rotation_angle.degrees')
            data_obj.meta_data.set('rotation_angle', rotation_angle)
        x = \
            data_obj.backing_file['entry1/instrument/lab_sxy/lab_sx'][()]
        data_obj.meta_data.set('x', x)
        # axis label

        ### set the y
        y = \
            data_obj.backing_file['entry1/instrument/lab_sxy/lab_sy'][()]
        pos = np.zeros((2,len(y[0])))
        pos[0,:] = x[0]
        pos[1,:] = y[0]
        data_obj.meta_data.set('xy', pos)
        # axis label
        labels.append('xy.microns')
        labels.append('idx.idx')
#         labels.append('y.microns')
        # now set them
        labels.extend(['spectra.counts'])
#         print "the labels are:"+str(labels)
        data_obj.set_axis_labels(*tuple(labels))


        dims = list(range(len(data_obj.get_shape())))
        spec_core = (-1,) # it will always be this
#         print spec_core

        spec_slice = tuple(dims[:-1])
#         print spec_slice
        logging.debug("is a spectrum")
        logging.debug("the spectrum cores are:"+str(spec_core))
        logging.debug("the spectrum slices are:"+str(spec_slice))
        data_obj.add_pattern("SPECTRUM", core_dims=spec_core,
                             slice_dims=spec_slice)
        logging.debug("the spectrum slices are:"+str(spec_slice))
        data_obj.add_pattern("SPECTRUM_STACK", core_dims=spec_core[:-1],
                             slice_dims=(-2,)+spec_slice)



        positions_label = (data_obj.get_data_dimension_by_axis_label('xy', contains=True),)
        rotation_label = (data_obj.get_data_dimension_by_axis_label('rotation_angle', contains=True),)

        data_obj.add_pattern("PROJECTION", core_dims=positions_label, slice_dims=tuple(set(dims)-set(positions_label)))
        sino_cores = rotation_label + positions_label
        data_obj.add_pattern("SINOGRAM", core_dims=sino_cores, slice_dims = tuple(set(dims)-set(sino_cores)))


        data_obj.add_pattern("4D_SCAN", core_dims=tuple(set(dims)-set(rotation_label)),
                        slice_dims=rotation_label)

#         data_obj.add_pattern("SINOGRAM", core_dir=(0,1),
#                      slice_dir=(2,3))
#         data_obj.add_pattern("PROJECTION", core_dir=(0,3),
#                      slice_dir=(1,2))
        self.set_data_reduction_params(data_obj)



# -*- coding: utf-8 -*-
# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_correction
   :platform: Unix
   :synopsis: A base class for dark and flat field corrections

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.plugin import Plugin


class BaseCorrection(Plugin):
    """
    A base class for dark and flat field correction plugins.

    """

    def __init__(self, name='BaseCorrection'):
        super(BaseCorrection, self).__init__(name)

    def setup(self):
        """
        Initial setup of all datasets required as input and output to the
        plugin.  This method is called before the process method in the plugin
        chain.
        """
        # set up the output dataset that is created by the plugin
        in_dataset, out_dataset = self.get_datasets()
        # copy all required information from in_dataset[0]
        out_dataset[0].create_dataset(in_dataset[0])
        # set information relating to the plugin data
        in_pData, out_pData = self.get_plugin_datasets()
        # set pattern_name and nframes to process for all datasets
        if 'pattern' in list(self.parameters.keys()):
            pattern = self.parameters['pattern']
        else:
            pattern = 'PROJECTION'

        in_pData[0].plugin_data_setup(pattern, self.get_max_frames())
        out_pData[0].plugin_data_setup(pattern, self.get_max_frames())

    def get_max_frames(self):
        return 'multiple'

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 1



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: completer
   :platform: Unix
   :synopsis: completer class for the configurator

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import os
import re

from savu.plugins import utils as pu
if os.name == 'nt':
    from . import win_readline as readline
else:
    import readline

RE_SPACE = re.compile('.*\s+$', re.M)


class Completer(object):

    def __init__(self, commands, plugin_list=[]):
        self.commands = commands
        self.plugin_list = plugin_list
        self.matches = None

    def _listdir(self, root):
        "List directory 'root' appending the path separator to subdirs."
        res = []
        for name in os.listdir(root):
            path = os.path.join(root, name)
            if os.path.isdir(path):
                name += os.sep
            res.append(name)
        return res

    def _complete_path(self, path=None):
        "Perform completion of filesystem path."
        if not path:
            return self._listdir('.')
        dirname, rest = os.path.split(path)
        tmp = dirname if dirname else '.'
        res = [os.path.join(dirname, p)
               for p in self._listdir(tmp) if p.startswith(rest)]
        # more than one match, or single match which does not exist (typo)
        if len(res) > 1 or not os.path.exists(path):
            return res
        # resolved to a single directory, so return list of files below it
        if os.path.isdir(path):
            return [os.path.join(path, p) for p in self._listdir(path)]
        # exact file match terminates this completion
        return [path + ' ']

    def path_complete(self, args):
        if not args:
            return self._complete_path('.')
        # treat the last arg as a path and complete it
        return self._complete_path(args[-1])

    def complete_open(self, args):
        "Completions for the open commands."
        return self.path_complete(args)

    def complete_save(self, args):
        "Completions for the save commands."
        return self.path_complete(args)

    def complete_list(self, args):
        "Completions for the list commands."
        list_args = [c.strip() for c in self._get_collections()]
        list_args += self.plugin_list
        if not args[0]:
            return list_args
        return [x for x in list_args if x.lower().startswith(args[0].lower())]

    def complete_add(self, args):
        "Completions for the add commands."
        list_args = self.plugin_list
        if not args[0] or len(args) is 2:
            return list_args
        return [x for x in list_args if x.lower().startswith(args[0].lower())]

    def _get_collections(self):
        """ Get plugin collection names. """
        import savu.plugins as plugins
        import copy

        path = plugins.__path__[0]
        exclude_dir = ['driver', 'utils']
        arrow = ' ==> '
        for root, dirs, files in os.walk(path):
            depth = root.count(os.path.sep) - path.count(os.path.sep)
            dirs[:] = [d for d in dirs if d not in exclude_dir]
            if depth == 0:
                colls = copy.copy(dirs)
            else:
                sep = '' if depth == 1 else ' ' + arrow if depth == 2 else \
                    (depth-2)*6*' ' + arrow
                pos = colls.index(sep + os.path.basename(root)) + 1
                sep = ' ' + arrow if depth == 1 else (depth-1)*6*' ' + arrow
                for i in range(len(dirs)):
                    colls.insert(pos, sep + dirs[i])
                    pos += 1
        return colls

    def complete_params(self, args):
        if not args[0]:
            return list(pu.plugins.keys())
        return [x for x in list(pu.plugins.keys()) if x.startswith(args[0])]

    def complete(self, text, state):
        "Generic readline completion entry point."
        if state == 0:
            self.matches = self.__get_matches(readline.get_line_buffer())
        return self.matches[state]

    def __get_matches(self, read_buffer):
        # show all commands
        line = readline.get_line_buffer().split()
        if not line:
            return [c + ' ' for c in list(self.commands.keys())]
        else:
            # account for last argument ending in a space
            if RE_SPACE.match(read_buffer):
                line.append('')

            # resolve command to the implementation function
            cmd = line[0].strip()
            if cmd in list(self.commands.keys()):
                impl = getattr(self, 'complete_%s' % cmd)
                args = line[1:]
                if args:
                    return (impl(args) + [None])
                return [cmd + ' ']
            return [c + ' ' for c in list(self.commands.keys()) if
                    c.startswith(cmd)] + [None]



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Remove stripe artefacts
   :platform: Unix
   :synopsis: A plugin working in sinogram space to remove large stripe artefacts
.. moduleauthor:: Nghia Vo <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
from savu.data.plugin_list import CitationInformation
import numpy as np
from scipy.ndimage import median_filter
from scipy.ndimage import binary_dilation
from scipy.ndimage import uniform_filter1d
from scipy import interpolate

@register_plugin
class RemoveUnresponsiveAndFluctuatingRings(Plugin, CpuPlugin):
    """

    Method to remove unresponsive and fluctuating stripe artefacts in a\
    sinogram (<-> ring artefacts in a reconstructed image). 

    :param size: Size of the median filter window. Greater is stronger\
    . Default: 71.
    :param snr: Ratio used to detect locations of stripes. Greater is\
     less sensitive. Default: 3.0.
    """

    def __init__(self):
        super(RemoveUnresponsiveAndFluctuatingRings, self).__init__(
                "RemoveUnresponsiveAndFluctuatingRings")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0])
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', 'single')
        out_pData[0].plugin_data_setup('SINOGRAM', 'single')

    def detect_stripe(self, listdata, snr):
        """
        Algorithm 4 in the paper. Used to locate stripe positions.
        ---------
        Parameters: - listdata: 1D normalized array.
                    - snr: ratio used to discriminate between useful
                        information and noise.
        ---------
        Return:     - 1D binary mask.
        """
        numdata = len(listdata)
        listsorted = np.sort(listdata)[::-1]
        xlist = np.arange(0, numdata, 1.0)
        ndrop = np.int16(0.25 * numdata)
        (_slope, _intercept) = np.polyfit(
            xlist[ndrop:-ndrop-1], listsorted[ndrop:-ndrop - 1], 1)
        numt1 = _intercept + _slope * xlist[-1]
        noiselevel = np.abs(numt1 - _intercept)
        val1 = np.abs(listsorted[0] - _intercept) / noiselevel
        val2 = np.abs(listsorted[-1] - numt1) / noiselevel
        listmask = np.zeros_like(listdata)
        if (val1 >= snr):
            upper_thresh = _intercept + noiselevel * snr * 0.5
            listmask[listdata > upper_thresh] = 1.0
        if (val2 >= snr):
            lower_thresh = numt1 - noiselevel * snr * 0.5
            listmask[listdata <= lower_thresh] = 1.0
        return listmask
    
    def remove_large_stripe(self, matindex, sinogram, snr, size):
        """
        Algorithm 5 in the paper. Use to remove residual stripes
        ---------
        Parameters: - sinogram: 2D array.
                    - snr: ratio used to discriminate between useful
                        information and noise.
                    - size: window size of the median filter.
        ---------
        Return:     - stripe-removed sinogram.
        """
        badpixelratio = 0.05
        (nrow, ncol) = sinogram.shape
        ndrop = np.int16(badpixelratio * nrow)
        sinosorted = np.sort(sinogram, axis=0)
        sinosmoothed = median_filter(sinosorted, (1, size))
        list1 = np.mean(sinosorted[ndrop:nrow - ndrop], axis=0)
        list2 = np.mean(sinosmoothed[ndrop:nrow - ndrop], axis=0)
        listfact = list1 / list2
        listmask = self.detect_stripe(listfact, snr)
        listmask = binary_dilation(listmask, iterations=1).astype(listmask.dtype)
        matfact = np.tile(listfact,(nrow,1))
        sinogram = sinogram / matfact
        sinogram1 = np.transpose(sinogram)
        matcombine = np.asarray(np.dstack((matindex, sinogram1)))
        matsort = np.asarray(
            [row[row[:, 1].argsort()] for row in matcombine])
        matsort[:, :, 1] = np.transpose(sinosmoothed)
        matsortback = np.asarray(
            [row[row[:, 0].argsort()] for row in matsort])
        sino_corrected = np.transpose(matsortback[:, :, 1])
        listxmiss = np.where(listmask > 0.0)[0]
        sinogram[:, listxmiss] = sino_corrected[:, listxmiss]
        return sinogram

    def pre_process(self):
        in_pData = self.get_plugin_in_datasets()
        width_dim = \
            in_pData[0].get_data_dimension_by_axis_label('detector_x')
        height_dim = \
            in_pData[0].get_data_dimension_by_axis_label('rotation_angle')
        sino_shape = list(in_pData[0].get_shape())
        self.width1 = sino_shape[width_dim]
        self.height1 = sino_shape[height_dim]
        listindex = np.arange(0.0, self.height1, 1.0)
        self.matindex = np.tile(listindex,(self.width1,1))        
        self.size = np.clip(np.int16(self.parameters['size']), 1, self.width1-1)
        self.snr = np.clip(np.float32(self.parameters['snr']), 1.0, None)
        
    def process_frames(self, data):
        """
        Algorithm 6 and 5 in the paper. Remove unresponsive and fluctuating\
        stripes using algorithm 6. Then using algorithm 5 to clean residual\
        stripes.
        """
        sinogram = np.copy(data[0])        
        sinosmoothed = np.apply_along_axis(uniform_filter1d, 0, sinogram, 10)
        listdiff = np.sum(np.abs(sinogram - sinosmoothed), axis=0)
        nmean = np.mean(listdiff)
        listdiffbck = median_filter(listdiff, self.size)
        listdiffbck[listdiffbck == 0.0] = nmean
        listfact = listdiff / listdiffbck
        listmask = self.detect_stripe(listfact, self.snr)
        listmask = binary_dilation(listmask, iterations=1).astype(listmask.dtype)
        listmask[0:2] = 0.0
        listmask[-2:] = 0.0
        listx = np.where(listmask < 1.0)[0]
        listy = np.arange(self.height1)
        matz = sinogram[:, listx]
        finter = interpolate.interp2d(listx, listy, matz, kind='linear')
        listxmiss = np.where(listmask > 0.0)[0]
        if len(listxmiss) > 0:
            matzmiss = finter(listxmiss, listy)
            sinogram[:, listxmiss] = matzmiss
        # Use algorithm 5 to remove residual stripes
        sinogram = self.remove_large_stripe(self.matindex, sinogram, self.snr, self.size) 
        return sinogram

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The code of ring removal is the implementation of the work of \
            Nghia T. Vo et al. taken from algorithm 4, 5, 6 in this paper.")
        cite_info.bibtex = \
            ("@article{Vo:18,\n" +
             "title={Superior techniques for eliminating ring artifacts in\
              X-ray micro-tomography},\n" +
             "author={Nghia T. Vo, Robert C. Atwood,\
              and Michael Drakopoulos},\n" +
             "journal={Opt. Express},\n" +
             "volume={26},\n" +
             "number={22},\n" +
             "pages={28396--28412},\n" +
             "year={2018},\n" +
             "publisher={OSA}" +
             "}")
        cite_info.doi = "doi: DOI: 10.1364/OE.26.028396"
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: find_peaks
   :platform: Unix
   :synopsis: A plugin to find the peaks

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
from savu.plugins.filters.base_filter import BaseFilter
import peakutils as pe
import numpy as np
from scipy.signal import savgol_filter


@register_plugin
class FindPeaks(BaseFilter, CpuPlugin):
    """
    This plugin uses peakutils to find peaks in spectra. This is then metadata.
    :param out_datasets: Create a list of the dataset(s). Default: ['Peaks'].
    :param thresh: Threshold for peak detection Default: 0.03.
    :param min_distance: Minimum distance for peak detection. Default: 15.
    """

    def __init__(self):
        super(FindPeaks, self).__init__("FindPeaks")

    def process_frames(self, data):
        data = data[0]
        # filter to smooth noise
        data = savgol_filter(data, 51, 3)
        PeakIndex = pe.indexes(data, thres=self.parameters['thresh'],
                               min_dist=self.parameters['min_distance'])

        PeakIndexOut = np.zeros(data.shape)
        PeakIndexOut[PeakIndex] = 1
        return PeakIndexOut

    def setup(self):
        # set up the output dataset that is created by the plugin
        in_dataset, out_dataset = self.get_datasets()
        # set information relating to the plugin data
        in_pData, out_pData = self.get_plugin_datasets()

        # set pattern_name and nframes to process for all datasets
        in_pData[0].plugin_data_setup("SPECTRUM", self.get_max_frames())
        nFrames = in_pData[0].get_total_frames()
        labels = ['frames.index', 'position.idx']
        shape = (nFrames, in_dataset[0].get_shape()[-1])
        out_dataset[0].create_dataset(axis_labels=labels, shape=shape)
        out_dataset[0].add_pattern("CHANNEL", slice_dims=(1,), core_dims=(0,))
        out_dataset[0].add_pattern("SPECTRUM", slice_dims=(0,), core_dims=(1,))
        out_pData[0].plugin_data_setup("SPECTRUM", self.get_max_frames())

    def get_max_frames(self):
        return 'single'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: A wrapper for TomoPhantom software
   :platform: Unix
   :synopsis: TomoPhantom package provides an access to simulated phantom libraries and projection data 2D/3D

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

import savu.plugins.utils as pu
from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

import tomophantom
from tomophantom import TomoP2D, TomoP3D
from tomophantom.supp.artifacts import _Artifacts_
import os
from savu.data.plugin_list import CitationInformation
import numpy as np

@register_plugin
class TomoPhantom(Plugin, CpuPlugin):
    """
    A plugin for TomoPhantom software which generates synthetic phantoms and \
    projection data (2D from Phantom2DLibrary.dat and 3D from Phantom3DLibrary.dat)

    :param geom_model: Select a model (integer) from the library (see TomoPhantom files). Default: 1.
    :param geom_model_size: Set the size of the phantom. Default: 256.
    :param geom_projections_total: The total number of projections. Default: 360.
    :param geom_detectors_horiz: The size of _horizontal_ detectors. Default: 300.
    :param artifacts_noise_type: Set the noise type, Poisson or Gaussian. Default: 'Poisson'.
    :param artifacts_noise_sigma: Define noise amplitude. Default: 5000.
    :param artifacts_misalignment_maxamplitude: Incorporate misalignment into projections (in pixels). Default: None.
    :param artifacts_zingers_percentage: add broken pixels to projections, e.g. 0.25. Default: None.
    :param artifacts_stripes_percentage: the amount of stripes in the data, e.g. 1.0. Default: None.
    :param artifacts_stripes_maxthickness: defines the maximal thickness of a stripe. Default: 3.0.
    :param artifacts_stripes_intensity: to incorporate the change of intensity in the stripe. Default: 0.3.
    :param artifacts_stripes_type: set the stripe type between full and partial. Default: 'full'.
    :param artifacts_stripes_variability: the intensity variability of a stripe. Default:  0.007.
    :param out_datasets: Default out dataset names. Default: ['tomo', 'model']
    """

    def __init__(self):
        super(TomoPhantom, self).__init__('TomoPhantom')

    def get_max_frames(self):
        return 'single'

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 2

    def map_volume_dimensions(self, data):
        data._finalise_patterns()
        dim_rotAngle = data.get_data_patterns()['PROJECTION']['main_dir']
        sinogram = data.get_data_patterns()['SINOGRAM']
        dim_detY = sinogram['main_dir']

        core_dirs = sinogram['core_dims']
        dim_detX = list(set(core_dirs).difference(set((dim_rotAngle,))))[0]

        dim_volX = dim_rotAngle
        dim_volY = dim_detY
        dim_volZ = dim_detX
        return dim_volX, dim_volY, dim_volZ

    def setup(self):
        in_dataset, self.out_dataset = self.get_datasets()
        in_pData, self.out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', self.get_max_frames())

        [self.out_shape_sino, out_shape_phantom] = self.new_shape(in_dataset[0].get_shape(), in_dataset[0])
        self.out_dataset[0].create_dataset(patterns=in_dataset[0],
                                           axis_labels=in_dataset[0],
                                           shape=self.out_shape_sino)
        self.out_pData[0].plugin_data_setup('SINOGRAM',self.get_max_frames())

        dim_volX, dim_volY, dim_volZ = \
            self.map_volume_dimensions(in_dataset[0])

        axis_labels = [0]*3
        axis_labels = {in_dataset[0]:
                       [str(dim_volX) + '.voxel_x.voxels',
                        str(dim_volY) + '.voxel_y.voxels',
                        str(dim_volZ) + '.voxel_z.voxels']}

        self.out_dataset[1].create_dataset(axis_labels=axis_labels,
                                           shape=out_shape_phantom)
        self.out_dataset[1].add_volume_patterns(dim_volX, dim_volY, dim_volZ)

        self.out_pData[1].plugin_data_setup('VOLUME_XZ', self.get_max_frames())

        self.angles = np.linspace(0.0,179.999,self.parameters['geom_projections_total'],dtype='float32')
        self.out_dataset[0].meta_data.set('rotation_angle', self.angles)

    def new_shape(self, full_shape, data):
        # calculate a new output data shape based on the input data shape
        core_dirs = data.get_core_dimensions()
        new_shape_sino = list(full_shape)
        new_shape_phantom = list(full_shape)
        new_shape_sino= (self.parameters['geom_projections_total'], new_shape_sino[1], self.parameters['geom_detectors_horiz'])
        for dim in core_dirs:
            new_shape_phantom[dim] = self.parameters['geom_model_size']
        return [tuple(new_shape_sino), tuple(new_shape_phantom)]

    def pre_process(self):
        # set parameters for TomoPhantom:
        self.model = self.parameters['geom_model']
        self.dims = self.parameters['geom_model_size']
        self.proj_num = self.parameters['geom_projections_total']
        self.detectors_num = self.parameters['geom_detectors_horiz']
        path = os.path.dirname(tomophantom.__file__)
        self.path_library2D = os.path.join(path, "Phantom2DLibrary.dat")
        self.path_library3D = os.path.join(path, "Phantom3DLibrary.dat")
        #print "The full data shape is", self.get_in_datasets()[0].get_shape()

    def process_frames(self, data):
        # print "The input data shape is", data[0].shape
        if (self.out_shape_sino[1] == 1):
            # create a 2D phantom
            model = TomoP2D.Model(self.model, self.dims, self.path_library2D)
            # create a 2D sinogram
            projdata_clean = TomoP2D.ModelSino(self.model, self.dims, self.detectors_num, self.angles, self.path_library2D)
            # Adding artifacts and noise
            # forming dictionaries with artifact types
            _noise_ =  {'type' : self.parameters['artifacts_noise_type'],
                        'sigma' : self.parameters['artifacts_noise_sigma'],
                        'seed' : 0,
                        'prelog' : False}

            # misalignment dictionary
            _sinoshifts_ = {}
            if self.parameters['artifacts_misalignment_maxamplitude'] is not None:
                _sinoshifts_ = {'maxamplitude' : self.parameters['artifacts_misalignment_maxamplitude']}

            # adding zingers and stripes
            _zingers_ = {}
            if self.parameters['artifacts_zingers_percentage'] is not None:
                _zingers_ = {'percentage' : self.parameters['artifacts_zingers_percentage'],
                            'modulus' : 10}

            _stripes_ = {}
            if self.parameters['artifacts_stripes_percentage'] is not None:
                _stripes_ = {'percentage' : self.parameters['artifacts_stripes_percentage'],
                             'maxthickness' : self.parameters['artifacts_stripes_maxthickness'],
                             'intensity' : self.parameters['artifacts_stripes_intensity'],
                             'type' : self.parameters['artifacts_stripes_type'],
                             'variability' : self.parameters['artifacts_stripes_variability']}

            if self.parameters['artifacts_misalignment_maxamplitude'] is not None:
                [projdata, shifts] = _Artifacts_(projdata_clean, _noise_, _zingers_, _stripes_, _sinoshifts_)
            else:
                projdata = _Artifacts_(projdata_clean, _noise_, _zingers_, _stripes_, _sinoshifts_)
        else:
            # create a 3D phantom
            frame_idx = self.out_pData[0].get_current_frame_idx()[0]
            model = TomoP3D.ModelSub(self.model, self.dims, (frame_idx, frame_idx+1), self.path_library3D)
            model = np.swapaxes(model,0,1)
            model = np.flipud(model[:,0,:])
            # create a 3D projection data
            projdata_clean = TomoP3D.ModelSinoSub(self.model, self.dims, self.detectors_num, self.dims, (frame_idx, frame_idx+1), self.angles, self.path_library3D)
            # Adding artifacts and noise
            # forming dictionaries with artifact types
            _noise_ =  {'type' : self.parameters['artifacts_noise_type'],
                        'sigma' : self.parameters['artifacts_noise_sigma'],
                        'seed' : 0,
                        'prelog' : False}

            # misalignment dictionary
            _sinoshifts_ = {}
            if self.parameters['artifacts_misalignment_maxamplitude'] is not None:
                _sinoshifts_ = {'maxamplitude' : self.parameters['artifacts_misalignment_maxamplitude']}

            # adding zingers and stripes
            _zingers_ = {}
            if self.parameters['artifacts_zingers_percentage'] is not None:
                _zingers_ = {'percentage' : self.parameters['artifacts_zingers_percentage'],
                            'modulus' : 10}

            _stripes_ = {}
            if self.parameters['artifacts_stripes_percentage'] is not None:
                _stripes_ = {'percentage' : self.parameters['artifacts_stripes_percentage'],
                             'maxthickness' : self.parameters['artifacts_stripes_maxthickness'],
                             'intensity' : self.parameters['artifacts_stripes_intensity'],
                             'type' : self.parameters['artifacts_stripes_type'],
                             'variability' : self.parameters['artifacts_stripes_variability']}

            if self.parameters['artifacts_misalignment_maxamplitude'] is not None:
                [projdata, shifts] = _Artifacts_(projdata_clean, _noise_, _zingers_, _stripes_, _sinoshifts_)
            else:
                projdata = _Artifacts_(projdata_clean, _noise_, _zingers_, _stripes_, _sinoshifts_)
            projdata = np.swapaxes(projdata,0,1)
        return [projdata, model]

    def get_citation_information(self):
        cite_info1 = CitationInformation()
        cite_info1.name = 'citation1'
        cite_info1.description = \
            ("TomoPhantom is a software package to generate 2D-4D analytical phantoms and their Radon transforms for various testing purposes.")
        cite_info1.bibtex = \
            ("@article{kazantsevTP2018,\n" +
             "title={TomoPhantom, a software package to generate 2D-4D analytical phantoms for CT image reconstruction algorithm benchmarks},\n" +
             "author={Daniil and Kazantsev, Valery and Pickalov, Srikanth and Nagella, Edoardo and Pasca, Philip and Withers},\n" +
             "journal={Software X},\n" +
             "volume={7},\n" +
             "number={--},\n" +
             "pages={150--155},\n" +
             "year={2018},\n" +
             "publisher={Elsevier}\n" +
             "}")
        cite_info1.endnote = \
            ("%0 Journal Article\n" +
             "%T TomoPhantom, a software package to generate 2D-4D analytical phantoms for CT image reconstruction algorithm benchmarks\n" +
             "%A Kazantsev, Daniil\n" +
             "%A Pickalov, Valery\n" +
             "%A Nagella, Srikanth\n" +
             "%A Pasca, Edoardo\n" +
             "%A Withers, Philip\n" +
             "%J Software X\n" +
             "%V 7\n" +
             "%N --\n" +
             "%P 150--155\n" +
             "%@ --\n" +
             "%D 2018\n" +
             "%I Elsevier\n")
        cite_info1.doi = "doi:10.1016/j.softx.2018.05.003"
        return [cite_info1]



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: monitor_correction
   :platform: Unix
   :synopsis: A plugin to crop a spectra

.. moduleauthor:: Aaron D. Parsons <scientificsoftware@diamond.ac.uk>
"""

import logging
from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin


@register_plugin
class MonitorCorrection(Plugin, CpuPlugin):
    """
     corrects the data to the monitor counts.
     This plugin corrects data[0] from data[1] by dividing. We allow a scale \
     and offset due to I18's uncalibrated ic

    :param in_datasets: A list of the dataset(s) to \
        process. Default: ['to_be_corrected','monitor'].
    :param nominator_scale: a. Default: 1.0.
    :param nominator_offset: b. Default: 0.0.
    :param denominator_scale: c. Default: 1.0.
    :param denominator_offset: d. Default: 0.0.
    :param pattern: the pattern to apply it to. Default:"PROJECTION".
    """

    def __init__(self):
        logging.debug("correcting data")
        super(MonitorCorrection, self).__init__("MonitorCorrection")

    def process_frames(self, data):
        monitor = data[1]
        to_be_corrected = data[0]
        nom_scale = self.parameters['nominator_scale']
        denom_scale = self.parameters['denominator_scale']
        nom_off = self.parameters['nominator_offset']
        denom_off = self.parameters['denominator_offset']
        monitor = monitor * denom_scale + denom_off
        to_be_corrected = to_be_corrected * nom_scale + nom_off
        out = to_be_corrected / monitor
        return out

    def setup(self):
        in_datasets, out_datasets = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        tobecorrected = in_datasets[0]
        monitor = in_datasets[1]
        corrected = out_datasets[0]
        corrected.create_dataset(tobecorrected)
        in_pData, out_pData = self.get_plugin_datasets()
#         pattern = tobecorrected.get_data_patterns().keys()[-1]# just take the first one
        frames = self.get_max_frames()
        in_pData[0].plugin_data_setup(self.parameters['pattern'], frames)
        in_pData[1].plugin_data_setup(self.parameters['pattern'], frames)
        out_pData[0].plugin_data_setup(self.parameters['pattern'], frames)

    def get_max_frames(self):
        return 'multiple'
        
    def nOutput_datasets(self):
        return 1

    def nInput_datasets(self):
        return 2




# Copyright 2018 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: stage_motion
   :platform: Unix
   :synopsis: Calculate stage motion from motor positions.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import copy
import numpy as np
import itertools
import time
import logging

from typing import Union

from savu.plugins.plugin import Plugin
from savu.plugins.utils import register_plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.core.utils import ensure_string

from pmacparser.pmac_parser import PMACParser


@register_plugin
class StageMotion(Plugin, CpuPlugin):
    """
    A Plugin to calculate stage motion from motion positions.
    :param in_datasets: Create a list of the \
        dataset(s). Default: ["pmean"].
    :param out_datasets: Create a list of the \
        dataset(s). Default: ["qmean"].
    :param use_min_max: Also use the min and max datasets \
        including all combinations of min, mean and max. Default: False.
    :param extra_in_datasets: The extra datasets \
        to use as input for min and max. Default: ["pmin", "pmax"].
    :param extra_out_datasets: The extra datasets \
        to use as output for min and max. Default: ["qmin", "qmax"].
    """

    NUM_OUTPUT_Q_VARS = 9
    MEAN_INDEX = 0
    NUM_DATASETS = 1

    def __init__(self):
        super(StageMotion, self).__init__("StageMotion")
        self.parser = None
        self.variables = {}
        self.pvals = None
        self.num_datasets = self.NUM_DATASETS
        self.use_min_max = False

    def _check_parameters(self):
        use_min_max = self.parameters['use_min_max']
        logging.debug("Config use min max: " + str(use_min_max))
        logging.debug("Experiment meta data use min max: " + \
                          str(self.exp.meta_data.get('use_minmax')))

        # Check the parameters from the config file and the experiment data
        # to check whether to use extra datasets

        # Config value takes priority, check if None
        if self.parameters['use_min_max'] is None:
            # Config was None, so check the value in the experiment data file
            if self.exp.meta_data.get('use_minmax'):
                self.use_min_max = True
        elif self.parameters['use_min_max']:
            self.use_min_max = True

    def pre_process(self):
        # Check the parameters from the config file and the experiment data
        if self.parameters['use_min_max'] is None:
            # Config was None, so check the value in the experiment data file
            if self.exp.meta_data.get('use_minmax'):
                self.use_min_max = True
        elif self.parameters['use_min_max']:
            self.use_min_max = True

        # Get the output shape from the plugin
        out_pdata = self.get_plugin_out_datasets()[0]
        self.out_pshape = out_pdata.get_shape()

        # Get the kinematic program to run and the static variables
        program_from_data = self.get_in_meta_data()[0].get_dictionary()['program']
        variables_from_data = self.get_in_meta_data()[0].get_dictionary()['variables']

        # Create the list of code lines from the kinematic program input
        code_lines = []
        for line in program_from_data:
            code_lines.append(ensure_string(line))

        # Set the static variables from the meta input
        for i in range(len(variables_from_data)):
            name = ensure_string(variables_from_data['Name'][i])
            self.variables[name] = variables_from_data['Value'][i]

        # Create the PMACParser instance
        self.parser = PMACParser(code_lines)

        # Find which dimension of the data passed to process_frames has axis
        # label 'motor'
        p_data = self.get_plugin_in_datasets()[0]
        self.motor_dim = p_data.get_data_dimension_by_axis_label('motor')

        # Fet the axis label values associated with the axis label 'motor'
        data = self.get_in_datasets()[0]
        self.pvals = data.meta_data.get('motor')

    def process_frames(self, data):
        # Data coming in is either an array of 1 or 3 elements
        # If 1 element, then just MEAN data, if 3 then MEAN, MIN, MAX
        # The last dimension represents the P values for each slice
        # Returns either 1 or 3 datasets of size to match the input

        mean_output = np.zeros(self.out_pshape)

        # Check if using all combinations of min, mean and max
        if self.use_min_max:
            # Using min, mean and max
            min_output = np.zeros(self.out_pshape)
            max_output = np.zeros(self.out_pshape)

            num_frames, num_p = np.array(data[0]).shape

            # Create a new data array consisting of all permutations of the P values
            # for min, mean and max, in order to call the parser only once

            # Create an index of the min, mean and max permutations
            index = np.array([list(itertools.product([0, 1, 2], repeat=num_p))])
            index = index.reshape(index.shape[0] * index.shape[1], index.shape[2])

            # Stack and transpose the data array to get it so that the last two dimensions
            # represent the min,mean,max and then the P value
            new_array = np.dstack(np.transpose(data))

            # Create the new data array by slicing the new array with the permutations
            new_data = new_array[:, index, np.arange(num_p)]

            # Now, set the variables to pass to the parser with the new data array
            for index, pval in enumerate(self.pvals):
                self.variables['P' + str(pval)] = new_data[:, :, index]

            # Run the parser
            parse_result = self.parser.parse(self.variables)

            # Extract the Q values from the parse results
            for i in range(1, self.NUM_OUTPUT_Q_VARS + 1):
                if 'Q' + str(i) in parse_result:
                    q_result = parse_result['Q' + str(i)]

                    if not hasattr(q_result, 'size'):
                        size = 1
                    else:
                        size = q_result.size

                    if size > 1:
                        mean_output[i - 1] = q_result[:, self.MEAN_INDEX]
                        max_output[i - 1] = np.amax(q_result, axis=1)
                        min_output[i - 1] = np.amin(q_result, axis=1)
                    else:
                        mean_output[i - 1] = q_result
                        max_output[i - 1] = q_result
                        min_output[i - 1] = q_result

            return [mean_output, min_output, max_output]

        else:
            # Only using mean data
            data_mean = data[self.MEAN_INDEX]

            # Run the parser on the mean input data to get the mean output data
            for index, pval in enumerate(self.pvals):
                self.variables['P' + str(pval)] = data_mean[:, index]

            # Run the parser
            output_vars = self.parser.parse(self.variables)

            # Extract the Q values from the parse results
            for i in range(1, self.NUM_OUTPUT_Q_VARS + 1):
                if 'Q' + str(i) in output_vars:
                    mean_output[i - 1] = output_vars['Q' + str(i)]

            return [mean_output]

    def post_process(self):
        pass

    def setup(self):
        # take in n x m
        # return n x m x r

        in_datasets, out_datasets = self.get_datasets()
        self.motor_data_dim = \
            in_datasets[0].get_data_dimension_by_axis_label('motor')

        # additional dimension information
        dim_len = self.NUM_OUTPUT_Q_VARS  # = r
        dim_pos = 0
        dim_axis_name = 'Qvalues'
        dim_axis_unit = 'units'

        new_axis = '.'.join(['~' + str(dim_pos), dim_axis_name, dim_axis_unit])
        axis_labels = {in_datasets[0]: [str(self.motor_data_dim), new_axis]}
        self.shape = self.__get_output_shape(in_datasets[0], dim_pos, dim_len)
        pattern = self.__update_pattern_information(in_datasets[0], dim_pos)

        # Populate the output datasets with required information
        number_of_datasets_in_use = 1
        if self.parameters['use_min_max'] is None:
            # Config was None, so check the value in the experiment data file
            if self.exp.meta_data.get('use_minmax'):
                self.use_min_max = True
        elif self.parameters['use_min_max']:
            self.use_min_max = True
        if self.use_min_max:
            number_of_datasets_in_use = 3

        logging.debug("Number of datasets: " + str(number_of_datasets_in_use))

        for i in range(number_of_datasets_in_use):
            out_datasets[i].create_dataset(axis_labels=axis_labels,
                                           shape=self.shape)
            out_datasets[i].add_pattern('MOTOR_POSITION', **pattern)

        # ================== populate plugin datasets =========================
        in_pData, out_pData = self.get_plugin_datasets()
        for i in range(number_of_datasets_in_use):
            in_pData[i].plugin_data_setup('MOTOR_POSITION', 'multiple')
            out_pData[i].plugin_data_setup('MOTOR_POSITION', 'multiple')
        # =====================================================================

    def __update_pattern_information(self, dObj, dim_pos):
        pattern = copy.deepcopy(dObj.get_data_patterns()['MOTOR_POSITION'])
        pattern['core_dims'] = (dim_pos,)
        slice_dims = pattern['slice_dims']
        slice_dims = [i if i < dim_pos else i + 1 for i in list(slice_dims)]
        pattern['slice_dims'] = tuple(slice_dims)
        return pattern

    def __get_output_shape(self, dObj, dim_pos, dim_len):
        shape = list(dObj.get_shape())
        del shape[self.motor_data_dim]
        shape.insert(dim_pos, dim_len)
        return tuple(shape)

    def nInput_datasets(self):
        self._check_parameters()
        # If using extra datasets, set them up and increase the number of datasets
        if self.use_min_max:
            logging.debug("Using min and max datasets")
            self.num_datasets += len(self.parameters['extra_in_datasets'])
            self.parameters['in_datasets'].extend(
                    self.parameters['extra_in_datasets'])
        return self.num_datasets

    def nOutput_datasets(self):
        if self.use_min_max:
            self.parameters['out_datasets'].extend(
                    self.parameters['extra_out_datasets'])
        return self.num_datasets



# Copyright 2020 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: tomo_phantom_quantification
   :platform: Unix
   :synopsis: A Tomophantom plugin to calculate quantitative values e.g. MSE, RMSE, SSIM

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

import numpy as np
import tomophantom
from tomophantom.supp.qualitymetrics import QualityTools

@register_plugin
class TomoPhantomQuantification(Plugin, CpuPlugin):
    """
    A plugin to calculate some standard image quality metrics

    :param out_datasets: The default names . Default: ['quantification_values'].
    """

    def __init__(self):
        super(TomoPhantomQuantification, self).__init__("TomoPhantomQuantification")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('VOLUME_XZ', 'single')
        in_pData[1].plugin_data_setup('VOLUME_XZ', 'single')

        fullData = in_dataset[0]
        slice_dirs = list(in_dataset[0].get_slice_dimensions())
        self.new_shape = (np.prod(np.array(fullData.get_shape())[slice_dirs]), 1)
        out_dataset[0].create_dataset(shape=self.new_shape,
                                      axis_labels=['quantval', 'value'],
                                      remove=True,
                                      transport='hdf5')
        out_dataset[0].add_pattern("METADATA", core_dims=(1,), slice_dims=(0,))
        out_pData[0].plugin_data_setup('METADATA', self.get_max_frames())

    def process_frames(self, data):
        data_temp1 = data[0]
        data_temp2 = data[1]
        indices = np.where(np.isnan(data_temp1))
        data_temp1[indices] = 0.0
        indices = np.where(np.isnan(data_temp2))
        data_temp2[indices] = 0.0

        # collecting values for each slice
        Qtools = QualityTools(data_temp1, data_temp2)
        RMSE = Qtools.rmse()
        print("The Root Mean Square Error is {}".format(RMSE))
        slice_values = [RMSE]
        return np.array([slice_values])

    def nInput_datasets(self):
        return 2
    def nOutput_datasets(self):
        return 1
    def get_max_frames(self):
        return 'single'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: mrc_loader
   :platform: Unix
   :synopsis: A class for MRC data.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import h5py
import tempfile
import numpy as np

from savu.plugins.loaders.base_loader import BaseLoader
from savu.plugins.utils import register_plugin
from savu.data.data_structures.data_types.mrc import MRC as MrcType


@register_plugin
class MrcLoader(BaseLoader):
    """
    Load Medical Research Council (MRC) formatted image data.
    :param angles: A python statement to be evaluated \
    (e.g np.linspace(0, 180, nAngles)) or a file. Default: None.
    :param name: The name assigned to the dataset. Default: 'tomo'.

    """

    def __init__(self, name='MrcLoader'):
        super(MrcLoader, self).__init__(name)

    def setup(self):
        exp = self.exp
        data_obj = exp.create_data_object('in_data', self.parameters['name'])

        filename = exp.meta_data.get("data_file")
        data_obj.data = MrcType(data_obj, filename)

        # dummy file
        path = exp.meta_data.get("data_file")
        filename = path.split('/')[-1] + '.h5'
        data_obj.backing_file = \
            h5py.File(tempfile.mkdtemp() + '/' + filename, 'a')

        self._set_rotation_angles(data_obj)

        nDims = len(data_obj.data.get_shape())
        if nDims is 3:
            rot = 0
            detY = 1
            detX = 2
            data_obj.set_axis_labels('rotation_angle.degrees',
                                     'detector_y.pixel',
                                     'detector_x.pixel')

            data_obj.add_pattern('PROJECTION', core_dims=(detX, detY),
                                 slice_dims=(rot,))
            data_obj.add_pattern('SINOGRAM', core_dims=(detX, rot),
                                 slice_dims=(detY,))
        elif nDims is 4:
            # 4D patterns need to be set here
            pass
        else:
            raise Exception('Incorrect number of data dimensions found in '
                            'mrc loader')

        data_obj.set_shape(data_obj.data.get_shape())
        return data_obj

    def _set_rotation_angles(self, data_obj):
        angles = self.parameters['angles']

        if angles is None:
            angles = np.linspace(0, 180, data_obj.data.get_shape()[0])
        else:
            try:
                angles = eval(angles)
            except:
                try:
                    angles = np.loadtxt(angles)
                except:
                    raise Exception('Cannot set angles in loader.')

        n_angles = len(angles)
        data_angles = data_obj.data.get_shape()[0]
        if data_angles != n_angles:
            raise Exception("The number of angles %s does not match the data "
                            "dimension length %s", n_angles, data_angles)
        data_obj.meta_data.set("rotation_angle", angles)




# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_image_saver
   :platform: Unix
   :synopsis: Handles common functionality for savers that save separate image files.

.. moduleauthor:: Dan Nixon <daniel.nixon@stfc.ac.uk>

"""

from mpi4py import MPI
import os

from savu.plugins.savers.base_saver import BaseSaver
from savu.plugins.utils import register_plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin


class BaseImageSaver(BaseSaver, CpuPlugin):
    """
    """

    def __init__(self, name):
        super(BaseImageSaver, self).__init__(name)
        self.count = None
        self.folder = None
        self.data_name = None
        self.file_name = None
        self.group_name = None
        self.max_files = 100000

    def pre_process(self):
        self.data_name = self.get_in_datasets()[0].get_name()
        self.count = 0
        self.group_name = self._get_group_name(self.data_name)
        self.folder = "%s/%s-%s" % (self.exp.meta_data.get("out_path"),
                                    self.name, self.data_name)
        if self.parameters['prefix']:
            self.filename = "%s/%s" % (self.folder, self.parameters['prefix'])
        else:
            self.filename = "%s/%s_" % (self.folder, self.data_name)
            self.filename += '%s_' % self.exp.meta_data.get("datafile_name")

        if MPI.COMM_WORLD.rank == 0:
            if not os.path.exists(self.folder):
                os.makedirs(self.folder)

    def setup(self):
        super(BaseImageSaver, self).setup()
        in_pData = self.get_plugin_in_datasets()[0]
        if in_pData.get_total_frames() > self.max_files:
            emsg = "Sorry, your data is too big to use an image saver."
            raise Exception(emsg)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: tomopy_recon
   :platform: Unix
   :synopsis: Wrapper around the tomopy reconstruction algorithms.  See \
       'http://tomopy.readthedocs.io/en/latest/api/tomopy.recon.algorithm.html'

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.driver.cpu_plugin import CpuPlugin

# import tomopy before numpy
import logging
logger = logging.getLogger()
level = logger.getEffectiveLevel()
logger.setLevel(30)
import tomopy
logger.setLevel(level)
import numpy as np

from savu.data.plugin_list import CitationInformation
from savu.plugins.utils import register_plugin
from savu.plugins.reconstructions.base_recon import BaseRecon


@register_plugin
class TomopyRecon(BaseRecon, CpuPlugin):
    """
     A wrapper to the tomopy reconstruction library. Extra keywords not \
     required for the chosen algorithm will be ignored.

    :u*param algorithm: The reconstruction algorithm (art|bart|fbp|gridrec|\
        mlem|osem|ospml_hybrid|ospml_quad|pml_hybrid|pml_quad\
        |sirt). Default: 'gridrec'.
    :u*param filter_name: Valid for fbp|gridrec, options: none|shepp|cosine|\
     hann|hamming|ramlak|parzen|butterworth). Default: 'ramlak'.
    :u*param reg_par: Regularization parameter for smoothing, valid for \
        ospml_hybrid|ospml_quad|pml_hybrid|pml_quad. Default: 0.0.
    :param n_iterations: Number of iterations - only valid for iterative \
    algorithms. Default: 1.
    :~param init_vol: Not an option. Default: None.
    :~param centre_pad: Not an option. Default: None.
    """

    def __init__(self):
        super(TomopyRecon, self).__init__("TomopyRecon")

    def pre_process(self):
        self.sl = self.get_plugin_in_datasets()[0].get_slice_dimension()
        vol_shape = self.get_vol_shape()
        filter_name = self.parameters['filter_name']
        # for backwards compatibility
        filter_name = 'none' if filter_name == None else filter_name
        options = {'filter_name': filter_name,
                   'reg_par': self.parameters['reg_par'] - 1,
                   'n_iterations': self.parameters['n_iterations'],
                   'num_gridx': vol_shape[0], 'num_gridy': vol_shape[2]}

        l = vol_shape[0]
        c = np.linspace(-l / 2.0, l / 2.0, l)
        x, y = np.meshgrid(c, c)
        ratio = self.parameters['ratio']
        if isinstance(ratio, list) or isinstance(ratio, tuple):
            self.ratio_mask = ratio[0]
            outer_mask = ratio[1]
            if isinstance(outer_mask, str):
                outer_mask = np.nan
        else:
            self.ratio_mask = ratio
            outer_mask = 0.0

        if isinstance(self.ratio_mask, float):
            r = (l - 1) * self.ratio_mask
            self.manual_mask = \
                np.array((x**2 + y**2 >= (r / 2.0)**2), dtype=np.float)
            self.manual_mask[self.manual_mask == 1] = outer_mask

        self.alg_keys = self.get_allowed_kwargs()
        self.alg = self.parameters['algorithm']
        self.kwargs = {key: options[key] for key in self.alg_keys[self.alg] if
                       key in options.keys()}
        self._finalise_data = self._transpose if self.parameters['outer_pad']\
            else self._apply_mask

    def process_frames(self, data):
        self.sino = data[0]
        self.cors, angles, vol_shape, init = self.get_frame_params()
        if init:
            self.kwargs['init_recon'] = init
        recon = tomopy.recon(self.sino, np.deg2rad(angles),
                             center=self.cors[0], ncore=1, algorithm=self.alg,
                             **self.kwargs)
        return self._finalise_data(recon)

    def _apply_mask(self, recon):
        if isinstance(self.ratio_mask, float):
            recon = tomopy.circ_mask(recon, axis=0, ratio=self.ratio_mask)
            recon = recon[0] + self.manual_mask
            recon = np.expand_dims(recon, 0)
        return self._transpose(recon)

    def _transpose(self, recon):
        return np.transpose(recon, (1, 0, 2))

    def get_max_frames(self):
        return 'multiple'

    def get_allowed_kwargs(self):
        return {
            'art': ['num_gridx', 'num_gridy', 'num_iter'],
            'bart': ['num_gridx', 'num_gridy', 'num_iter', 'num_block',
                     'ind_block'],
            'fbp': ['num_gridx', 'num_gridy', 'filter_name', 'filter_par'],
            'gridrec': ['num_gridx', 'num_gridy', 'filter_name', 'filter_par'],
            'mlem': ['num_gridx', 'num_gridy', 'num_iter'],
            'osem': ['num_gridx', 'num_gridy', 'num_iter', 'num_block',
                     'ind_block'],
            'ospml_hybrid': ['num_gridx', 'num_gridy', 'num_iter', 'reg_par',
                             'num_block', 'ind_block'],
            'ospml_quad': ['num_gridx', 'num_gridy', 'num_iter', 'reg_par',
                           'num_block', 'ind_block'],
            'pml_hybrid': ['num_gridx', 'num_gridy', 'num_iter', 'reg_par'],
            'pml_quad': ['num_gridx', 'num_gridy', 'num_iter', 'reg_par'],
            'sirt': ['num_gridx', 'num_gridy', 'num_iter'],
        }

    def get_padding_algorithms(self):
        """ A list of algorithms that allow the data to be padded. """
        return ['fbp', 'gridrec']

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("An algorithm from the TomoPy framework is used to perform the \
             reconstruction in this processing pipeline")
        cite_info.bibtex = \
            ("@article{gursoy2014tomopy,\n" +
             "title={TomoPy: a framework for the analysis of synchrotron,\n" +
             "tomographic data},\n" +
             "author={G{\"u}rsoy, Doga and De Carlo, Francesco and Xiao,\n" +
             "Xianghui and Jacobsen, Chris},\n" +
             "journal={Journal of synchrotron radiation},\n" +
             "volume={21},\n" +
             "number={5},\n" +
             "pages={1188--1193},\n" +
             "year={2014},\n" +
             "publisher={International Union of Crystallography}\n" +
             "}")
        cite_info.endnote = \
            ("%0 Journal Article\n" +
             "%T TomoPy: a framework for the analysis of synchrotron\n" +
             "tomographic data\n" +
             "%A Gursoy, Doga\n" +
             "%A De Carlo, Francesco" +
             "%A Xiao, Xianghui" +
             "%A Jacobsen, Chris" +
             "%J Journal of synchrotron radiation" +
             "%V 21" +
             "%N 5" +
             "%P 1188-1193" +
             "%@ 1600-5775" +
             "%D 2014" +
             "%I International Union of Crystallography")
        cite_info.doi = "doi: 10.1107/S1600577514013939"
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: vo_centering_test
   :platform: Unix
   :synopsis: unittest for vo centering routine

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import unittest
from savu.test import test_utils as tu
from savu.test.travis.framework_tests.plugin_runner_test import \
    run_protected_plugin_runner


class VoCenterTest(unittest.TestCase):

    def test_vo_centering(self):
        data_file = tu.get_test_data_path('24737.nxs')
        plist = 'vo_centering_test/vo_centering_test.nxs'
        process_file = tu.get_test_process_path(plist)
        run_protected_plugin_runner(tu.set_options(data_file,
                                                   process_file=process_file))

    def test_vo_centering_single_sino_loader(self):
        data_file = tu.get_test_data_path('24737.nxs')
        plist = 'vo_centering_test/vo_centering_single_sino_loader_test.nxs'
        process_file = tu.get_test_process_path(plist)
        run_protected_plugin_runner(tu.set_options(data_file,
                                                   process_file=process_file))

    def test_vo_centering_single_sino_centering(self):
        data_file = tu.get_test_data_path('24737.nxs')
        plist = 'vo_centering_test/vo_centering_single_sino_centering_test.nxs'
        process_file = tu.get_test_process_path(plist)
        run_protected_plugin_runner(tu.set_options(data_file,
                                                   process_file=process_file))

    def test_vo_centering_single_sinos(self):
        data_file = tu.get_test_data_path('24737.nxs')
        plist = 'vo_centering_test/vo_centering_single_sinos_test.nxs'
        process_file = tu.get_test_process_path(plist)
        run_protected_plugin_runner(tu.set_options(data_file,
                                                   process_file=process_file))

if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: poly_background_estimator
   :platform: Unix
   :synopsis: A plugin to find the peaks

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
from savu.plugins.filters.base_filter import BaseFilter
import numpy as np


@register_plugin
class PolyBackgroundEstimator(BaseFilter, CpuPlugin):
    """
    This plugin uses peakutils to find peaks in spectra. This is then metadata.
    :param out_datasets: Create a list of the dataset(s). Default: ['Peaks'].
    :param n: max number of polys. Default: 2.
    :param MaxIterations: max number of iterations. Default: 12.
    :param weights: weightings to apply. Default: '1/data'.
    :param pvalue: ratio of variance between successive poly \
        iterations. Default: 0.9.
    """

    def __init__(self):
        super(PolyBackgroundEstimator,
              self).__init__("PolyBackgroundEstimator")

    def process_frames(self, data):
        data = data[0]
        x = self.axis
        n = self.parameters['n']
        if self.parameters['weights'] == '1/data':
            weights = 1.0 / data
        elif self.parameters['weights'] == '1/data^2':
            weights = 1.0 / data**2
        pvalue = self.parameters['pvalue']
        MaxIterations = self.parameters['MaxIterations']
        zu, _c, _poly, _weight, _index = \
            self.poly_background_estimator(x, data, n, weights, MaxIterations,
                                           pvalue, fixed=True)

        #print "The shape of zu is:"+str(zu.shape)
        return zu

    def setup(self):
        in_dataset, out_datasets = self.get_datasets()
        out = out_datasets[0]
        out.create_dataset(in_dataset[0])
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SPECTRUM', self.get_max_frames())
        out_pData[0].plugin_data_setup('SPECTRUM', self.get_max_frames())
        in_meta = self.get_in_meta_data()[0]
        # get the axis
        alabel = \
            list(in_dataset[0].data_info.get('axis_labels')[-1].keys())[0]
        self.axis = in_meta.get(alabel)

    def get_max_frames(self):
        return 'single'

    def __generate_parameters(self, n, weight, xdata, ydata):
        """
        generates polynomials based on
        S. Steenstrup J. Appl. Cryst. (1981). 14, 226--229
        "A Simple Procedure for Fitting a Background to a Certain Class of \
        Measured Spectra". The polynomial parameters are based on weights \
        supplied as part of the fitting fit.
        """
        Npoints = xdata.size
        poly = np.zeros((n, Npoints), dtype=np.float64, order='F')
        gamma = np.zeros(n, dtype=np.float64, order='F')
        alpha = np.zeros(n, dtype=np.float64, order='F')
        beta = np.zeros(n, dtype=np.float64, order='F')
        a = np.zeros(n, dtype=np.float64, order='F')

        alpha[0] = (weight * xdata).sum() / weight.sum()
        beta[0] = 0.0
        poly[0] = 1.0
        poly[1] = xdata - alpha[0]
        for j in range(n):
            if j > 1:
                poly[j] = (xdata - alpha[j - 1]) * \
                    poly[j - 1] - beta[j - 1] * poly[j - 2]

            p = poly[j]
            g = (weight * p * p).sum()
            gamma[j] = g
            a[j] = (weight * ydata * p / g).sum()

            if j > 0:
                alpha[j] = (weight * xdata * p * p).sum() / g
                beta[j] = \
                    (weight * xdata * p * poly[j - 1]).sum() / gamma[j - 1]

        return alpha, gamma, beta, a, poly

    def poly_background_estimator(self, xdata, ydata, n=2, weights=None,
                                  maxIterations=12, pvalue=0.9, fixed=False):
        """
        Background estimator based on orthogonal polynomials

        Input:
        xdata,ydata (numpy arrays of same length)
        pvalue  :   ratio of variance in poly to poly value at which to stop.
        0.9 default

        Output:
                background,polynomial weights, polynomials

        S. Steenstrup J. Appl. Cryst. (1981). 14, 226--229
        "A Simple Procedure for Fitting a Background to a Certain Class of \
        Measured Spectra"

        """
        m = 0
        ud = False
        Npoints = xdata.size
        ydata = np.clip(ydata, 0.0001, ydata.max())
        weight = 1.0 / ydata
    #     zu = np.zeros(Npoints, dtype=np.float64)
        myiter = 0
        vartest = True
        c_old = []
        if(fixed):
            maxIterations = n+1
            n = 2
            myiter = 2
        AIC_OLD = 1.0e34
        while myiter < maxIterations:
            myiter += 1
            _alpha, gamma, _beta, c, poly = \
                self.__generate_parameters(n, weight, xdata, ydata)
            zu = (c[:, np.newaxis] * poly).sum(axis=0)
            y_diff = ydata - zu
            rss = (y_diff*y_diff).sum()
            Eu = (weight * y_diff * y_diff).sum()
            f = Npoints - n - m
            if (Eu < (f + m + np.sqrt(2.0 * f))):
                break
            sigma = np.sqrt(Eu / (f * gamma))
            m = 0.0
            index = ydata > (zu + 2.0 * np.sqrt(np.abs(zu)))
            weight = 1.0 / np.abs(zu)
            tw = ydata[index] - zu[index]
            weight[index] = 1.0 / (tw * tw)
            m = Npoints - index.sum()
            if(fixed):
                n += 1
            else:
                if myiter > 1:
                    clen = min(len(c_old), len(c))
                    vartest = True
                    for i in range(clen):
                        if((c[i]-sigma[i]) < c_old[i] and c_old[i] < (c[i] + sigma[i])):
                            vartest = True
                        else:
                            vartest = False
                            break
                if vartest:
                    if abs(sigma[n-1] / c[n-1]) < pvalue:
                        if ud:
                            break
                        n = n + 1
                    else:
                        if n > 3:
                            n = n - 1
                            ud = True
            c_old = c
        #print "finished", index.sum(), len(zu), Npoints, n, m
        return zu, c, poly, weight, index




# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: plugin_template3
   :platform: Unix
   :synopsis: A template to create a plugin that reduces the data dimensions.

.. moduleauthor:: Developer Name <email@address.ac.uk>

"""

import copy
import numpy as np

from savu.plugins.plugin import Plugin
from savu.plugins.utils import register_plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin


@register_plugin
class PluginTemplate7(Plugin, CpuPlugin):
    """
    A plugin template that increases the data dimensions.

    :param pattern: How to slice the data. Default: 'PROJECTION'.
    :param axis_len: Length of the new dimension. Default: 10.
    :param axis_label: Axis label for the new axis. Default: 'scan'.
    :param axis_unit: Axis unit for the new axis. Default: 'number'.
    """

    def __init__(self):
        super(PluginTemplate7, self).__init__('PluginTemplate7')

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 1

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()

        #=================== populate output dataset ==========================
        # Due to the increase in dimensions, the out_dataset will have
        # different axis_labels, patterns and shape to the in_dataset and
        # these will need to be defined.
        # For more information about the syntax used here see:
        # http://savu.readthedocs.io/en/latest/api_plugin/savu.data.data_structures.data_create

        # AMEND THE PATTERNS: The output dataset will have one dimension more
        # than the in_dataset, so add another slice dimensions the patterns
        add_dim = str(len(in_dataset[0].get_shape()))
        patterns = {in_dataset[0]: ['.'.join(['*', add_dim])]}

        # AMEND THE AXIS LABELS: Add an extra slice dim to all axis labels
        axis_list = ['.'.join(['~' + add_dim, self.parameters['axis_label'],
                            self.parameters['axis_unit']])]
        axis_labels = {in_dataset[0]: axis_list}

        # AMEND THE SHAPE: Remove the two unrequired dimensions from the
        # original shape and add a new dimension shape.
        self.rep = self.parameters['axis_len']
        shape = list(in_dataset[0].get_shape())
        shape.append(self.rep)

        # populate the output dataset
        out_dataset[0].create_dataset(
                patterns=patterns,
                axis_labels=axis_labels,
                shape=tuple(shape))

        slice_dim = (out_dataset[0].get_data_dimension_by_axis_label(
                'detector_y'),)
        core_dims = set(range(0, len(shape))).difference(set(slice_dim))
        sinomovie = {'core_dims': tuple(core_dims), 'slice_dims': slice_dim}
        out_dataset[0].add_pattern("SINOMOVIE", **sinomovie)
        print(out_dataset[0].get_data_patterns()['SINOMOVIE'])

        #================== populate plugin datasets ==========================
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('PROJECTION', 'single')
        out_pData[0].plugin_data_setup('PROJECTION', self.rep,
                 slice_axis=self.parameters['axis_label'])
        #======================================================================

    def pre_process(self):
        self.ndims = len(self.get_plugin_in_datasets()[0].get_shape())

    def process_frames(self, data):
        rep_mat = np.repeat(
                np.expand_dims(data[0], self.ndims), self.rep, axis=self.ndims)
        return rep_mat

    def post_process(self):
        pass


# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: nxxrd_loader
   :platform: Unix
   :synopsis: A class for loading nxxrd data

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""
import os.path as os
import h5py
import logging
import math
from savu.plugins.loaders.mapping_loaders.base_multi_modal_loader \
    import BaseMultiModalLoader
from savu.plugins.utils import register_plugin
import savu.test.test_utils as tu


@register_plugin
class NxxrdLoader(BaseMultiModalLoader):
    """
    A class to load tomography data from an NXxrd file

    :param calibration_path: path to the calibration file. Default: None.
    :param name: The name assigned to the dataset. Default: 'xrd'.
    """

    def __init__(self):
        super(NxxrdLoader, self).__init__("NxxrdLoader")
        # converting lengths to metres
        self.mm = 1e-3
        self.angstrom = 1e-10

    def setup(self):
        self.parameters['in_datasets'] = self.parameters['name']
        path = 'instrument/detector/data'
        data_obj, xrd_entry = \
            self.multi_modal_setup('NXxrd', path, self.parameters['name'])
        mono_energy = data_obj.backing_file[
            xrd_entry.name + '/instrument/monochromator/energy'][()]
        self.exp.meta_data.set("mono_energy", mono_energy)

        self._get_calibration_info(data_obj)
        self._add_diffraction_pattern(data_obj)
        self.set_data_reduction_params(data_obj)

    def _add_diffraction_pattern(self, dObj):
        detX = dObj.get_data_dimension_by_axis_label('detector_x')
        detY = dObj.get_data_dimension_by_axis_label('detector_y')
        cdims = (detX, detY)
        all_dims = list(range(len(dObj.get_shape())))
        sdims = tuple(set(all_dims).difference(cdims))
        dObj.add_pattern("DIFFRACTION", core_dims=cdims, slice_dims=sdims)

    def _get_calibration_info(self, data_obj):
        cfile = h5py.File(self.get_cal_path(), 'r')
        det_str = 'entry/instrument/detector'
        mData = data_obj.meta_data

        try:
            self._set_calibration_new(mData, det_str, cfile)
            logging.debug('.... its the version in DAWN 2.0')
        except KeyError:
            try:
                self._set_calibration_legacy(mData, det_str, cfile)
                logging.debug('.... its the legacy version pre-DAWN 2.0')
            except KeyError:
                emsg = "We don't know what type of calibration file this is"
                logging.warning(emsg)
        cfile.close()

    def _set_calibration_new(self, mData, det_str, cfile):
        xpix_entry = det_str + '/detector_module/fast_pixel_direction'
        xpix = cfile[xpix_entry][()]*self.mm
        mData.set("x_pixel_size", xpix)

        beam_center_x = cfile[det_str + '/beam_center_x'][()]*self.mm
        mData.set("beam_center_x", beam_center_x)

        beam_center_y = cfile[det_str + '/beam_center_y'][()]*self.mm
        mData.set("beam_center_y", beam_center_y)

        distance = cfile[det_str + '/distance'][()]*self.mm
        mData.set("distance", distance)

        wentry = '/entry1/calibration_sample/beam/incident_wavelength'
        wlength = cfile[wentry][()]*self.angstrom
        mData.set("incident_wavelength", wlength)

        yaw = -cfile[det_str + '/transformations/euler_b'][()]
        mData.set("yaw", yaw)

        roll = cfile[det_str + '/transformations/euler_c'][()]-180.0
        mData.set("roll", roll)

    def _set_calibration_legacy(self, mData, det_str, cfile):
        xpix = cfile[det_str + '/x_pixel_size'][()]*self.mm
        mData.set("x_pixel_size", xpix)

        beam_center_x = cfile[det_str + '/beam_center_x'][()]*xpix
        mData.set("beam_center_x", beam_center_x)

        beam_center_y = cfile[det_str + '/beam_center_y'][()]*xpix
        mData.set("beam_center_y", beam_center_y)

        distance = cfile[det_str + '/distance'][()]*self.mm
        mData.set("distance", distance)

        wl_entry = '/entry/calibration_sample/beam/incident_wavelength'
        wavelength = cfile[wl_entry][()]*self.angstrom
        mData.set("incident_wavelength", wavelength)

        orien = cfile[det_str + '/detector_orientation'][...].reshape((3, 3))
        yaw = math.degrees(-math.atan2(orien[2, 0], orien[2, 2]))
        mData.set("yaw", -yaw)

        roll = math.degrees(-math.atan2(orien[0, 1], orien[1, 1]))
        mData.set("roll", roll)

    def get_cal_path(self):
        path = self.parameters['calibration_path']
        if path.split(os.sep)[0] == 'Savu':
            path = tu.get_test_data_path(path.split('/test_data/data')[1])
        return path



# Copyright 2015 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: utils
   :platform: Unix
   :synopsis: Simple data utility methods
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import copy

# A dictionary of available patterns (and ranks, needed for dawn)
pattern_list = {"SINOGRAM": 2,
                "PROJECTION": 2,
                "VOLUME_YZ": 2,
                "VOLUME_XZ": 2,
                "VOLUME_XY": 2,
                "VOLUME_3D": 3,
                "SPECTRUM": 1,
                "DIFFRACTION": 2,
                "CHANNEL": 1,
                "SPECTRUM_STACK": 2,
                "PROJECTION_STACK": 3,
                "METADATA": 0,
                "4D_SCAN": 3,
                "TIMESERIES": 1,
                "MOTOR_POSITION": 2,
                "TANGENTOGRAM": 2,
                "SINOMOVIE": 3,
                "SINOGRAM_STACK": 3}


def _deepcopy_data_object(dObj, new_obj):
    """ Deepcopy data object, associating hdf5 objects that can not be copied.
    """
    new_obj.meta_data = dObj.meta_data
    new_obj.pattern_list = copy.deepcopy(dObj.pattern_list)
    new_obj.data_info = copy.deepcopy(dObj.data_info)
    new_obj.exp = dObj.exp
    new_obj._plugin_data_obj = dObj._plugin_data_obj
    new_obj.dtype = copy.deepcopy(dObj.dtype)
    new_obj.remove = copy.deepcopy(dObj.remove)
    new_obj.group_name = dObj.group_name
    new_obj.group = dObj.group
    new_obj.backing_file = dObj.backing_file
    new_obj.data = dObj.data
    new_obj.next_shape = copy.deepcopy(dObj.next_shape)
    new_obj.orig_shape = copy.deepcopy(dObj.orig_shape)
    new_obj.previous_pattern = copy.deepcopy(dObj.previous_pattern)
    new_obj._set_transport_data(dObj.data_info.get('transport'))
    return new_obj


def get_available_pattern_types():
    return list(pattern_list.keys())


def get_pattern_rank(pattern):
    return pattern_list[pattern]



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: gpu_plugin
   :platform: Unix
   :synopsis: The driver for GPU plugins.

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import os
import copy
import logging
import numpy as np
import pynvml as pv
from mpi4py import MPI
from itertools import chain

from savu.plugins.driver.plugin_driver import PluginDriver
from savu.plugins.driver.basic_driver import BasicDriver

_base = BasicDriver if os.environ['savu_mode'] == 'basic' else PluginDriver


class GpuPlugin(_base):
    """
    The driver for GPU plugins.
    """

    def __init__(self):
        super(GpuPlugin, self).__init__()

    def _run_plugin(self, exp, transport):
        expInfo = exp.meta_data
        processes = copy.copy(expInfo.get("processes"))
        process = expInfo.get("process")

        gpu_processes = [False] * len(processes)
        idx = [i for i in range(len(processes)) if 'GPU' in processes[i]]
        for i in idx:
            gpu_processes[i] = True

        # set only GPU processes
        new_processes = [i for i in processes if 'GPU' in i]
        if not new_processes:
            raise Exception("THERE ARE NO GPU PROCESSES!")
        expInfo.set('processes', new_processes)

        nNodes = new_processes.count(new_processes[0])

        ranks = [i for i, x in enumerate(gpu_processes) if x]
        idx = [i for i in range(len(ranks)) if new_processes[i] == 'GPU0']
        diff = np.diff(np.array(idx)) if len(idx) > 1 else 1
        split = np.max(diff) if not isinstance(diff, int) else len(ranks)
        split_ranks = [ranks[n:n + split] for n in range(0, len(ranks), split)]
        ranks = list(chain.from_iterable(zip(*split_ranks)))

        self.__create_new_communicator(ranks, exp, process)

        if gpu_processes[process]:
            expInfo.set('process', self.new_comm.Get_rank())
            GPU_index = self.__calculate_GPU_index(nNodes)
            logging.debug("Running the GPU process %i with GPU index %i",
                          self.new_comm.Get_rank(), GPU_index)
            self.parameters['GPU_index'] = GPU_index
            os.environ['CUDA_DEVICE'] = str(GPU_index)
            self._run_plugin_instances(transport, communicator=self.new_comm)
            self.__free_communicator()
            expInfo.set('process', MPI.COMM_WORLD.Get_rank())
        else:
            logging.info('Not a GPU process: Waiting...')

        self.exp._barrier()
        expInfo.set('processes', processes)
        return

    def __create_new_communicator(self, ranks, exp, process):
        self.group = MPI.COMM_WORLD.Get_group()
        self.new_group = MPI.Group.Incl(self.group, ranks)
        self.new_comm = MPI.COMM_WORLD.Create(self.new_group)
        self.exp._barrier()

    def __free_communicator(self):
        self.group.Free()
        self.new_group.Free()
        self.new_comm.Free()

    def __calculate_GPU_index(self, nNodes):
        pv.nvmlInit()
        nGPUs = int(pv.nvmlDeviceGetCount())
        rank = self.new_comm.Get_rank()
        return int(rank / nNodes) % nGPUs



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: hdf5_utils
   :platform: Unix
   :synopsis: A class containing savu specific hdf5 functions.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import os
import h5py
import logging
from mpi4py import MPI

from savu.data.chunking import Chunking
#from savu.data.data_structures.data_types.data_plus_darks_and_flats \
#    import NoImageKey
from savu.data.data_structures.data_types.base_type import BaseType


NX_CLASS = 'NX_class'


class Hdf5Utils(object):
    """
    A class to save tomography data to a hdf5 file
    """

    def __init__(self, exp):
        self.plugin = None
        self.info = MPI.Info.Create()
        self.exp = exp
        # Get MPI I/O settings from the Savu config file
        settings = self.exp.meta_data.get(['system_params', 'mpi-io_settings'])
        for key, value in settings.items():
            self.info.Set(key, value)

    def _open_backing_h5(self, filename, mode, comm=MPI.COMM_WORLD, mpi=True):
        """
        Create a h5 backend for output data
        """
        if mpi:
            msg = self.__class__.__name__ + "_open_backing_h5 %s" + filename
            self.exp._barrier(communicator=comm, msg=msg+'1')

        kwargs = {'driver': 'mpio', 'comm': comm, 'info': self.info}\
            if self.exp.meta_data.get('mpi') and mpi else {}

        backing_file = h5py.File(filename, mode, **kwargs)

        if mpi:
            self.exp._barrier(communicator=comm, msg=msg+'2')

        if backing_file is None:
            raise IOError("Failed to open the hdf5 file")
        return backing_file

    def _link_datafile_to_nexus_file(self, data):
        filename = self.exp.meta_data.get('nxs_filename')

        with h5py.File(filename, 'a') as nxs_file:
            # entry path in nexus file
            name = data.get_name()
            group_name = self.exp.meta_data.get(['group_name', name])
            link = self.exp.meta_data.get(['link_type', name])
            name = data.get_name(orig=True)
            nxs_entry = self.__add_nxs_entry(nxs_file, link, group_name, name)
            self.__add_nxs_data(nxs_file, nxs_entry, link, group_name, data)

    def __add_nxs_entry(self, nxs_file, link, group_name, name):
        nxs_entry = '/entry/' + link
        nxs_entry += '_' + name if link == 'final_result' else "/" + group_name
        nxs_entry = nxs_file[nxs_entry]
        nxs_entry.attrs['signal'] = 'data'
        return nxs_entry

    def __add_nxs_data(self, nxs_file, nxs_entry, link, group_name, data):
        data_entry = nxs_entry.name + '/data'
        # output file path
        h5file = data.backing_file.filename

        if link == 'input_data':
            dataset = self.__is_h5dataset(data)
            if dataset:
                nxs_file[data_entry] = \
                    h5py.ExternalLink(os.path.abspath(h5file), dataset.name)
        else:
            # entry path in output file path
            m_data = self.exp.meta_data.get
            if not (link == 'intermediate' and
                    m_data('inter_path') != m_data('out_path')):
                h5file = h5file.split(m_data('out_folder') + '/')[-1]
            nxs_file[data_entry] = \
                h5py.ExternalLink(h5file, group_name + '/data')

    def __is_h5dataset(self, data):
        if isinstance(data.data, h5py.Dataset):
            return data.data
        try:
            if isinstance(data.data.data, h5py.Dataset):
                return data.data.data
        except:
            return False

    def create_dataset_nofill(self, group, name:str, shape, dtype, chunks=None):
        spaceid = h5py.h5s.create_simple(shape)
        plist = h5py.h5p.create(h5py.h5p.DATASET_CREATE)
        plist.set_fill_time(h5py.h5d.FILL_TIME_NEVER)
        if chunks not in [None, []] and isinstance(chunks, tuple):
            plist.set_chunk(chunks)
        typeid = h5py.h5t.py_create(dtype)
        group_name = (group.name + '/' + name).encode("ascii")
        datasetid = h5py.h5d.create(
            group.file.id, group_name, typeid, spaceid, plist)
        data = h5py.Dataset(datasetid)
        return data

    def _create_entries(self, data, key:str, current_and_next):
        msg = self.__class__.__name__ + '_create_entries'
        self.exp._barrier(msg=msg+'1')

        expInfo = self.exp.meta_data
        group_name = expInfo.get(["group_name", key])
        data.data_info.set('group_name', group_name)
        try:
            group_name = group_name + '_' + data.name
        except AttributeError:
            pass

        self.exp._barrier(msg=msg+'2')
        group = data.backing_file.require_group(group_name)
        self.exp._barrier(msg=msg+'3')
        shape = data.get_shape()

        if 'data' in group:
            data.data = group['data']
        elif current_and_next is 0:
            logging.warning('Creating the dataset without chunks')
            data.data = group.create_dataset("data", shape, data.dtype)
        else:
            chunk_max = self.__set_optimal_hdf5_chunk_cache_size(data, group)
            chunking = Chunking(self.exp, current_and_next)
            chunks = chunking._calculate_chunking(shape, data.dtype,
                                                  chunk_max=chunk_max)

            self.exp._barrier(msg=msg+'4')
            data.data = self.create_dataset_nofill(
                    group, "data", shape, data.dtype, chunks=chunks)

        self.exp._barrier(msg=msg+'5')
        return group_name, group

    def __set_optimal_hdf5_chunk_cache_size(self, data, group):
        # calculate the number first
        # change cache properties
        propfaid = group.file.id.get_access_plist()
        settings = list(propfaid.get_cache())
        pdict = self.exp.meta_data.get('system_params')
        max_chunk_size = pdict['max_chunk_size']
        chunk_cache_size = pdict['chunk_cache_size']
        settings[2] *= chunk_cache_size
        propfaid.set_cache(*settings)
        return max_chunk_size * 1e6  # convert MB to bytes

    def _close_file(self, data):
        """
        Closes the backing file
        """
        if data.backing_file is not None:
            try:
                msg = self.__class__.__name__ + "_close_file" + \
                data.backing_file.filename
                self.exp._barrier(msg=msg)
                logging.debug("Attempting to close the file ")
                filename = data.backing_file.filename
                data.backing_file.close()
                logging.debug("File close successful: %s", filename)
                data.backing_file = None
                data.filename = filename # needed for tests
                self.exp._barrier(msg=msg)
            except:
                logging.debug("File close unsuccessful", filename)

    def _reopen_file(self, data, mode):
        filename = data.backing_file.filename
        self._close_file(data)
        logging.debug(
                "Re-opening the backing file %s in mode %s" % (filename, mode))
        data.backing_file = self._open_backing_h5(filename, mode)
        entry = list(data.backing_file.keys())[0] + '/data'

        if isinstance(data.data, BaseType):
            data.data.data = data.backing_file[entry]
        elif isinstance(data.data, h5py._hl.dataset.Dataset):
            data.data = data.backing_file[entry]
        else:
            raise Exception('Unable to re-open the hdf5 file - unknown'
                            ' datatype')



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: plugins_util_test
   :platform: Unix
   :synopsis: unittest test class for plugin utils

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import unittest

import savu
import os

from savu.plugins import utils as pu
from savu.plugins import docstring_parser as doc
from savu.plugins import plugin as test_plugin


class Test(unittest.TestCase):

    def testGetPlugin(self):
        mod = "savu.plugins.plugin"
        plugin = pu.load_class(mod)()
        self.assertEqual(plugin.__class__, test_plugin.Plugin,
                         "Failed to load the correct class")
        self.assertRaises(NotImplementedError,
                          plugin.process_frames, None)

    def testfind_args(self):
        mod = "savu.plugins.filters.denoising.denoise_bregman_filter"
        plugin = pu.load_class(mod)()
        params = doc.find_args(plugin)
        self.assertEqual(len(params['param']), 4)

    def test_get_plugin_external_path(self):
        savu_path = os.path.split(savu.__path__[0])[0]
        mod = os.path.join(
                savu_path, "plugin_examples", "example_median_filter.py")
        plugin = pu.load_class(mod)()
        self.assertEqual(plugin.name, "ExampleMedianFilter")

    def _add_to_plugins_path(self, add_paths):
        env = "SAVU_PLUGINS_PATH"
        path = os.environ[env] if env in os.environ.keys() else ""
        os.environ[env] = ':'.join([path, add_paths])
        
    def test_get_plugins_paths(self):
        n_paths = len(pu.get_plugins_paths())
        self._add_to_plugins_path('/tmp')
        paths = pu.get_plugins_paths()
        self.assertEqual(len(paths), n_paths+1)

    def test_get_plugins_paths2(self):
        n_paths = len(pu.get_plugins_paths())
        self._add_to_plugins_path("/tmp/:/dev/:/home/")
        paths = pu.get_plugins_paths()
        self.assertEqual(len(paths), n_paths+3)

    def test_get_plugins_path_and_load(self):
        savu_path = os.path.split(savu.__path__[0])[0]
        plugin_path = os.path.join(savu_path, "plugin_examples")
        os.environ["SAVU_PLUGINS_PATH"] = plugin_path
        pu.get_plugins_paths()
        plugin = pu.load_class("example_median_filter")()
        self.assertEqual(plugin.name, "ExampleMedianFilter")
        os.environ["SAVU_PLUGINS_PATH"] = ""

if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: display_formatter
   :platform: Unix
   :synopsis: Classes for formatting plugin list output in the configurator.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import textwrap

from colorama import Fore, Back, Style

WIDTH = 85


class DisplayFormatter(object):

    def __init__(self, plugin_list):
        self.plugin_list_inst = plugin_list
        self.plugin_list = plugin_list.plugin_list

    def _get_string(self, **kwargs):
        out_string = []
        verbosity = kwargs.get('verbose', False)
        level = kwargs.get('level', 'user')

        start = kwargs.get('start', 0)
        stop = kwargs.get('stop', len(self.plugin_list))
        if stop == -1:
            stop = len(self.plugin_list)

        count = start
        plugin_list = self.plugin_list[start:stop]

        line_break = ('%s' % ('-'*WIDTH))
        out_string.append(line_break)
        for p_dict in plugin_list:
            count += 1
            description = \
                self._get_description(WIDTH, level, p_dict, count, verbosity)
            out_string.append(description)
            out_string.append(line_break)
        return '\n'.join(out_string)

    def _get_description(self, width, level, p_dict, count, verbose):
        if verbose == '-q':
            return self._get_quiet(p_dict, count, width)
        if not verbose:
            return self._get_default(level, p_dict, count, width)
        if verbose == '-v':
            return self._get_verbose(level, p_dict, count, width)
        if verbose == '-vv':
            return self._get_verbose_verbose(level, p_dict, count, width)

    def _get_plugin_title(self, p_dict, width, fore_colour, back_colour,
                          active="", quiet=False, pos=None):
        pos = "%2s" % (str(pos) + ")") if pos else ""
        title = "%s %s %s" % (active, pos, p_dict['name'])
        title = title if quiet else title+"(%s)" % p_dict['id']
        width -= len(title)
        return back_colour + fore_colour + title + " "*width + Style.RESET_ALL

    def _get_synopsis(self, plugin_name, width, colour_on, colour_off):
        doc_str = \
            self.plugin_list_inst._get_docstring_info(plugin_name)['synopsis']
        synopsis = \
            self._get_equal_lines(doc_str, width, colour_on, colour_off, " "*2)
        if not synopsis:
            return ''
        return "\n" + colour_on + synopsis + colour_off

    def _get_param_details(self, level, p_dict, width, desc=False):
        margin = 4
        keycount = 0
        joiner = "\n" + " "*margin
        params = ''

        dev_keys = [k for k in list(p_dict['data'].keys()) if k not in
                    p_dict['user'] + p_dict['hide']]
        keys = p_dict['user'] if level == 'user' else p_dict['user'] + dev_keys

        for key in keys:
            keycount += 1
            temp = "\n   %2i)   %20s : %s"
            # keycount = all_keys.index(key)+1
            params += temp % (keycount, key, p_dict['data'][key])
            if desc:
                pdesc = " ".join(desc[key].split())
                pdesc = joiner.join(textwrap.wrap(pdesc, width=width-margin))
                temp = '\n' + Fore.CYAN + ' '*margin + "%s" + Fore.RESET
                params += temp % pdesc
        return params

    def _get_extra_info(self, p_dict, width, colour_off, info_colour,
                        warn_colour):
        extra_info = self.plugin_list_inst._get_docstring_info(p_dict['name'])
        info = self._get_equal_lines(extra_info['info'], width, info_colour,
                                     colour_off, " "*2)
        warn = self._get_equal_lines(extra_info['warn'], width, warn_colour,
                                     colour_off, " "*2)
        info = "\n"+info if info else ''
        warn = "\n"+warn if warn else ''
        return info, warn

    def _get_equal_lines(self, string, width, colour_on, colour_off, offset):
        if not string or not colour_on:
            return ''
        string = str.splitlines(string)
        str_list = []
        for s in string:
            str_list += textwrap.wrap(s, width=width-len(offset))
        new_str_list = []
        for line in str_list:
            lwidth = width - len(line) - len(offset)
            new_str_list.append(
                colour_on + offset + line + " "*lwidth + colour_off)
        return "\n".join(new_str_list)


class DispDisplay(DisplayFormatter):

    def __init__(self, plugin_list):
        super(DispDisplay, self).__init__(plugin_list)

    def _get_quiet(self, p_dict, count, width, quiet=True):
        active = \
            '***OFF***' if 'active' in p_dict and not p_dict['active'] else ''
        p_dict['data'] = self._remove_quotes(p_dict['data'])
        pos = p_dict['pos'].strip() if 'pos' in list(p_dict.keys()) else count
        fore = Fore.RED + Style.DIM if active else Fore.LIGHTWHITE_EX
        back = Back.LIGHTBLACK_EX
        return self._get_plugin_title(p_dict, width, fore, back,
                                      active=active, quiet=quiet, pos=pos)

    def _get_default(self, level, p_dict, count, width):
        title = self._get_quiet(p_dict, count, width)
        params = self._get_param_details(level, p_dict, width)
        return title + params

    def _get_verbose(self, level, p_dict, count, width, breakdown=False):
        title = self._get_quiet(p_dict, count, width, quiet=False)
        colour_on = Back.LIGHTBLACK_EX + Fore.LIGHTWHITE_EX
        colour_off = Back.RESET + Fore.RESET
        synopsis = \
            self._get_synopsis(p_dict['name'], width, colour_on, colour_off)
        params = \
            self._get_param_details(level, p_dict, width, desc=p_dict['desc'])
        if breakdown:
            return title, synopsis, params
        return title + synopsis + params

    def _get_verbose_verbose(self, level, p_dict, count, width):
        title, synopsis, param_details = \
            self._get_verbose(level, p_dict, count, width, breakdown=True)
        info_c = Back.CYAN + Fore.LIGHTWHITE_EX
        warn_c = Back.WHITE + Fore.RED
        c_off = Back.RESET + Fore.RESET
        info, warn = self._get_extra_info(p_dict, width, c_off, info_c, warn_c)
        return title + synopsis + info + warn + param_details

    def _remove_quotes(self, data_dict):
        """ Remove quotes around variables for display
        """
        for key, val in data_dict.items():
            val = str(val).replace("'", "")
            data_dict[key] = val
        return data_dict

    def _notices(self):
        width = 86
        warnings = self.get_warnings(width)
        if warnings:
            notice = Back.RED + Fore.WHITE + "IMPORTANT PLUGIN NOTICES" +\
                Back.RESET + Fore.RESET + "\n"
            border = "*"*width + '\n'
            print((border + notice + warnings + '\n'+border))

    def get_warnings(self, width):
        # remove display styling outside of this class
        colour = Back.RESET + Fore.RESET
        warnings = []
        names = []
        for plugin in self.plugin_list:
            if plugin['name'] not in names:
                names.append(plugin['name'])
                warn = self.plugin_list_inst._get_docstring_info(
                        plugin['name'])['warn']
                if warn:
                    for w in warn.split('\n'):
                        string = plugin['name'] + ": " + w + '.'
                        warnings.append(self._get_equal_lines(
                            string, width-1, colour, colour, " "*2))
        return "\n".join(
            ["*" + "\n ".join(w.split('\n')) for w in warnings if w])


class ListDisplay(DisplayFormatter):

    def __init__(self, plugin_list):
        super(ListDisplay, self).__init__(plugin_list)

    def _get_quiet(self, p_dict, count, width):
        return self._get_plugin_title(p_dict, width, Fore.RESET, Back.RESET,
                                      quiet=True)

    def _get_default(self, level, p_dict, count, width):
        title = self._get_quiet(p_dict, count, width)
        synopsis = \
            self._get_synopsis(p_dict['name'], width, Fore.CYAN, Fore.RESET)
        return title + synopsis

    def _get_verbose(self, level, p_dict, count, width, breakdown=False):
        default_str = self._get_default(level, p_dict, count, width)
        info_c = Fore.CYAN
        c_off = Back.RESET + Fore.RESET
        info, warn = self._get_extra_info(p_dict, width, c_off, info_c, info_c)
        return default_str + info

    def _get_verbose_verbose(self, level, p_dict, count, width):
        all_params = self._get_param_details('all', p_dict, 100)
        default_str = self._get_default(level, p_dict, count, width)
        info_c = Fore.CYAN
        warn_c = Fore.RED
        c_off = Back.RESET + Fore.RESET
        info, warn = self._get_extra_info(p_dict, width, c_off, info_c, warn_c)
        return default_str + info + warn + all_params



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Remove stripe artefacts  
   :platform: Unix
   :synopsis: A plugin working in sinogram space to remove stripe artefacts     
.. moduleauthor:: Nghia Vo <scientificsoftware@diamond.ac.uk>    
   
"""
from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
from savu.data.plugin_list import CitationInformation
import numpy as np


@register_plugin
class RingRemovalRegularization(Plugin, CpuPlugin):
    """

    Method to remove stripe artefacts in a sinogram (<-> ring artefacts in a\ 
    reconstructed image) using a regularization-based method. 
    A simple improvement to handle partial stripes is included.    
    :param alpha: The correction strength. Smaller is stronger. Default: 0.005
    :param number_of_chunks: Divide the sinogram to many chunks of rows. \
    Default: 1
    """

    def __init__(self):
        super(RingRemovalRegularization, self).__init__(
            "RingRemovalRegularization")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0])
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', 'single')
        out_pData[0].plugin_data_setup('SINOGRAM', 'single')

    def pre_process(self):
        in_pData = self.get_plugin_in_datasets()
        width_dim = in_pData[0].get_data_dimension_by_axis_label('detector_x')
        height_dim = in_pData[0].get_data_dimension_by_axis_label(
            'rotation_angle')
        sino_shape = list(in_pData[0].get_shape())
        self.width1 = sino_shape[width_dim]
        self.height1 = sino_shape[height_dim]
        alpha = self.parameters['alpha']
        tau = 2.0 * np.arcsinh(np.sqrt(alpha) * 0.5)
        ilist = np.arange(0, self.width1)
        jlist = np.arange(0, self.width1)
        matjj, matii = np.meshgrid(jlist, ilist)
        mat1 = np.abs(matii - matjj)
        mat2 = matii + matjj
        mat1a = np.cosh((self.width1 - 1 - mat1) * tau)
        mat2a = np.cosh((self.width1 - mat2) * tau)
        self.mat_coe = -(np.tanh(0.5 * tau) /
                         (alpha * np.sinh(self.width1 * tau))) * (mat1a + mat2a)

    def process_frames(self, data):
        sinogram = np.copy(data[0])
        num_mean = np.mean(sinogram)
        sinogram[sinogram <= 0.0] = num_mean  # Note performance
        sinogram = -np.log(sinogram)
        num_chunks = np.clip(
            np.int16(self.parameters['number_of_chunks']), 1, self.height1)
        list_pos = np.array_split(np.arange(self.height1), num_chunks)
        list_grad = np.zeros(self.width1, dtype=np.float32)
        mat_grad = np.zeros((self.width1, self.width1), dtype=np.float32)
        for pos in list_pos:
            bindex = pos[0]
            eindex = pos[-1] + 1
            list_mean = np.mean(sinogram[bindex:eindex, :], axis=0)
            list_grad[1:-1] = - np.diff(list_mean, 2)
            list_grad[0] = list_mean[0] - list_mean[1]
            list_grad[-1] = list_mean[-1] - list_mean[-2]
            mat_grad[:] = list_grad
            list_corr = np.sum(mat_grad * self.mat_coe, axis=1)
            mat_corr = np.zeros(
                (eindex - bindex, self.width1), dtype=np.float32)
            mat_corr[:] = list_corr
            sinogram[bindex:eindex, :] = sinogram[bindex:eindex, :] + mat_corr
        return np.exp(-sinogram)

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The code of ring removal is the implementation of the work of \
            Sofya Titarenko et al. taken from this paper.")
        cite_info.bibtex = \
            ("@article{TITARENKO20101489,\n" +
             "title={An analytical formula for ring artefact suppression in \
             X-ray tomography},\n" +
             "author={Sofya Titarenko, Philip J. Withers and \
             Anatoly Yagola},\n" +
             "journal={Applied Mathematics Letters},\n" +
             "volume={23},\n" +
             "number={12},\n" +
             "pages={1489 - 1495},\n" +
             "year={2010},\n" +
             "publisher={Elsevier}" +
             "}")
        cite_info.doi = "doi: DOI: 10.1016/j.aml.2010.08.022"
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: plugin_template3
   :platform: Unix
   :synopsis: A template to create a plugin that reduces the data dimensions.

.. moduleauthor:: Developer Name <email@address.ac.uk>

"""

import copy
import numpy as np

from savu.plugins.plugin import Plugin
from savu.plugins.utils import register_plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin


@register_plugin
class PluginTemplate3(Plugin, CpuPlugin):
    """
    A plugin template that reduces the data dimensions, e.g. azimuthal \
    integration.

    :param num_bins: Length of the new dimension. Default: 10.
    """

    def __init__(self):
        super(PluginTemplate3, self).__init__('PluginTemplate3')

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 1

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()

        #=================== populate output dataset ==========================
        # Due to the reduction in dimensions, the out_dataset will have
        # different axis_labels, patterns and shape to the in_dataset and
        # these will need to be defined.
        # For more information about the syntax used here see:
        # http://savu.readthedocs.io/en/latest/api_plugin/savu.data.data_structures.data_create

        # AMEND THE PATTERNS: The output dataset will have one dimension less
        # than the in_dataset, so remove the final slice dimension from any
        # patterns you want to keep.
        rm_dim = str(in_dataset[0].get_data_patterns()
                     ['SINOGRAM']['slice_dims'][-1])
        patterns = ['SINOGRAM.' + rm_dim, 'PROJECTION.' + rm_dim]

        # AMEND THE AXIS LABELS: Find the dimensions to remove using their
        # axis_labels to ensure the plugin is as generic as possible and will
        # work for data in all orientations.
        axis_labels = copy.copy(in_dataset[0].get_axis_labels())
        rm_labels = ['detector_x', 'detector_y']
        rm_dims = sorted([in_dataset[0].get_data_dimension_by_axis_label(a)
                          for a in rm_labels])[::-1]
        for d in rm_dims:
            del axis_labels[d]
        # Add a new axis label to the list
        axis_labels.append({'Q': 'Angstrom^-1'})

        # AMEND THE SHAPE: Remove the two unrequired dimensions from the
        # original shape and add a new dimension shape.
        shape = list(in_dataset[0].get_shape())
        for d in rm_dims:
            del shape[d]
        shape += (self.get_parameters('num_bins'),)

        # populate the output dataset
        out_dataset[0].create_dataset(
                patterns={in_dataset[0]: patterns},
                axis_labels=axis_labels,
                shape=tuple(shape))

        # ASSOCIATE AN EXTRA PATTERN WITH THE DATASET: SINOGRAM and PROJECTION
        # patterns are already asssociated with the output dataset, but add 
        # another one.
        spectrum = \
            {'core_dims': (-1,), 'slice_dims': tuple(range(len(shape)-1))}
        out_dataset[0].add_pattern("SPECTRUM", **spectrum)
        #======================================================================

        #================== populate plugin datasets ==========================
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('DIFFRACTION', 'single')
        out_pData[0].plugin_data_setup('SPECTRUM', 'single')
        #======================================================================

    def pre_process(self):
        pass

    def process_frames(self, data):
        # do some processing here
        return np.arange(self.parameters['num_bins'])

    def post_process(self):
        pass



# Copyright 2020 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: a dezinger filter (based on the median) from the Larix software
   :platform: Unix
   :synopsis: A 2D/3D median-based dezinger plugin to apply to any data
.. moduleauthor::Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""
from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

import numpy as np
from larix.methods.misc import MEDIAN_DEZING

@register_plugin
class Dezinger(Plugin, CpuPlugin):
    """
    A plugin to apply 2D/3D median-based dezinger. The 3D capability is enabled\
    through padding. Note that the kernel_size in 2D will be kernel_size x kernel_size
    and in 3D case kernel_size x kernel_size x kernel_size.
    
    :u*param kernel_size: Kernel size of the median filter. Default: 3.
    :param outlier_mu: Threshold for detecting outliers, greater is less \
    sensitive. If very small, dezinger acts like a median filter. Default: 1.0.
    :u*param dimension: dimensionality of the filter 2D/3D. Default: '3D'.
    :u*param pattern: pattern to apply this to. Default: "PROJECTION".

    :config_warn: The dezinger plugin should be applied to normalised data \
    (e.g. AFTER DarkFlatFieldCorrection)
    """

    def __init__(self):
        super(Dezinger, self).__init__("Dezinger")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0])
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup(self.parameters['pattern'], 'single')
        out_pData[0].plugin_data_setup(self.parameters['pattern'], 'single')

    def process_frames(self, data):
        input_temp = np.float32(data[0])
        indices = np.where(np.isnan(input_temp))
        input_temp[indices] = 0.0
        if (self.parameters['dimension'] == '3D'):
            if (self.parameters['pattern'] == 'VOLUME_XY'):
                input_temp =np.swapaxes(input_temp,0,2)
            if ((self.parameters['pattern'] == 'VOLUME_XZ') or (self.parameters['pattern'] == 'SINOGRAM')):
                input_temp =np.swapaxes(input_temp,0,1)
        result = MEDIAN_DEZING(input_temp.copy(order='C'), self.parameters['kernel_size'], self.parameters['outlier_mu'])
        if (self.parameters['dimension'] == '3D'):
            if (self.parameters['pattern'] == 'VOLUME_XY'):
                result =np.swapaxes(result,0,2)
            if ((self.parameters['pattern'] == 'VOLUME_XZ') or (self.parameters['pattern'] == 'SINOGRAM')):
                result =np.swapaxes(result,0,1)
        return result

    def set_filter_padding(self, in_data, out_data):
        if (self.parameters['dimension'] == '3D'):
            padding = (self.parameters['kernel_size']-1)/2
        else:
            padding = 0
        in_data[0].padding = {'pad_multi_frames': padding}
        out_data[0].padding = {'pad_multi_frames': padding}

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 1

    def get_plugin_pattern(self):
        return self.parameters['pattern']



# Copyright 2015 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: data_notes
   :platform: Unix
   :synopsis: A module containing extended doc strings for the data module.
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""


def _set_preview_note():
    """
        Each ``preview_list`` element should be of the form
        ``start:stop:step:chunk``, where ``stop``, ``step`` and ``chunk`` are
        optional (defaults: ``stop`` = ``start`` +  1, ``step`` = 1, ``chunk`` = 1)
        but must be given in that order.

        .. note::
            **start:stop[:step]**
                represents the set of indices specified by:

                >>> indices = range(start, stop[, step])

            For more information see :func:`range`

            **start:stop:step:chunk (chunk > 1)**
                represents the set of indices specified by:

                >>> a = np.tile(np.arange(start, stop, step), (chunk, 1))
                >>> b = np.transpose(np.tile(np.arange(chunk)-chunk // 2, \
(a.shape[1], 1)))
                >>> indices = np.ravel(np.transpose(a + b))

                Chunk indicates how many values to take around each value in
                ``range(start, stop, step)``.  It is only available for slicing
                dimensions.

                .. warning:: If any indices are out of range (or negative)
                    then the list is invalid. When chunk > 1, new start and
                    end values will be:

                    >>> new_start = start - int(chunk / 2)
                    >>> new_end = range(start, stop, step)[-1] + \
(step - int(chunk / 2))

        **accepted values**:
            Each entry is executed using :func:`eval` so simple formulas are\
            allowed and may contain the following keywords:

            * ``:`` is a simplification for 0:end:1:1 (all values)
            * ``mid`` is int(shape[dim] / 2)-1
            * ``end`` is shape[dim]
    """


def image_key():
    """
        This is a helper function to be used after :meth:`savu.data.\
data_structures.data_create.DataCreate.create_dataset`,

            >>> out_dataset[0].create_dataset(in_dataset[0])
            >>> out_dataset[0].trim_output_data(in_dataset[0], image_key=0)

        if in_dataset[0] is a plugin input dataset with an image_key and 0 is
        the data index.
    """


def _create():
    """
    .. note:: **Creating a dataset**
            Each new dataset requires the following information:

            * ``shape``
            * ``axis_labels``
            * ``patterns``

        This function can be used to setup the required information in one
        of two ways:

            1. Passing a ``Data`` object as the only argument: All required
            information is coped from this data object. For example,

                >>> out_dataset[0].create_dataset(in_dataset[0])

            2. Passing kwargs: ``shape`` and ``axis_labels`` are required
            (see above for other optional arguments). For example,

                >>> out_dataset[0].create_dataset(axis_labels=labels, \
shape=new_shape)

        .. warning:: If ``pattern`` keyword is not used, patterns must be added
            after  :meth:`~savu.data.data_structures.data_create.DataCreate.\
create_dataset` by calling :func:`~savu.data.data_structures.data.Data.\
add_pattern`.
    """


def _shape():
    """
    .. note::
        ``shape`` keyword argument
            Options to pass are:

            1. Data object: Copy shape from the Data object.

                >>> out_dataset[0].create_dataset(axis_labels=labels, \
shape=in_dataset[0])

            2. tuple: Define shape explicity.

                >>> out_dataset[0].create_dataset(axis_labels=labels, \
shape=(10, 20, 30))

    """


def axis_labels():
    """
    .. note::
        ``axis_labels`` keyword argument
            Options to pass are:

            1. Data object: Copy all labels from the Data object.

                >>> out_dataset[0].create_dataset(axis_labels=in_dataset[0], \
shape=new_shape)

            2. {Data_obj: list}: Copy labels from the Data object and then
            remove or insert.

                * To remove dimensions: list_entry = 'dim'. For example, to
                  remove the first and last axis_labels from the copied list:

                >>> out_dataset[0].create_dataset(axis_labels=\
{in_dataset[0]: ['1', '-1']), shape=new_shape})


                * To add/replace dimensions: list_entry = 'dim.name.unit'.

                >>> out_dataset[0].create_dataset(axis_labels={in_dataset[0]: \
['2.det_x.pixel', '3.det_y.pixel']}, shape=new_shape)

                * To insert dimensions: list_entry = '~dim.name.unit'.

                >>> out_dataset[0].create_dataset(axis_labels={in_dataset[0]: \
['~2.det_x.pixel', '~3.det_y.pixel']}, shape=new_shape)

                (or a combination, where each successive step is applied after
                the previous changes have been made.)

            3. list: Where each element is of the form 'name.unit'.

                >>> out_dataset[0].create_dataset(axis_labels=['rotation.deg',\
 'det_x.pixel', 'det_y.pixel'], shape=new_shape)

    """


def patterns():
    """
    .. note::
        ``patterns`` keyword argument
            Options to pass are:

            1. Data object: Copy all patterns from the Data object.

                >>> out_dataset[0].create_dataset(axis_labels=labels, \
shape=new_shape, patterns=in_dataset[0])

            2. {Data_obj: list}: Copy only the patterns given in the list
            from the Data object.

                * Copy the patterns: list_entry = 'name'

                >>> out_dataset[0].create_dataset(axis_labels=labels, \
shape=new_shape, patterns={in_dataset[0]: ['SINOGRAM', 'PROJECTION']})

                * Copy patterns but remove dimensions: list_entry = \
'name1.r1,r2...':

                >>> out_dataset[0].create_dataset(axis_labels=labels, \
shape=new_shape, patterns={in_dataset[0]: ['SINOGRAM.1', 'PROJECTION.1']})

                * Copy ALL patterns but remove dimensions: list_entry = \
'\*.r1,r2...':

                >>> out_dataset[0].create_dataset(axis_labels=labels, \
shape=new_shape, patterns={in_dataset[0]: '*.0'})

    """


def _padding():
    """
    - 'dim' should be replaced with the dimension to pad
    - 'pad' should be replaced with the amount to pad
    - 'before' and 'after' are optional positional keywords specifying \
        padding 'before' or 'after' the data for the specified dimension index\
        (if neither are specified BOTH will be padded)

    """



# Copyright 2015 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: data_notes
   :platform: Unix
   :synopsis: A module containing extended doc strings for the data module.
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""


def datasets_notes():
    """ Get {0} objects associated with the {1}_datasets.

    :returns: A list of all {0} objects associated with {1}_datasets for the
        current plugin.
    :rtype: list({0})
    """


def two_datasets_notes():
    """ Get {0} objects associated with in_datasets and
    out_datasets.

    :returns: Two lists of all {0} objects associated with
        in_datasets and out_datasets respectively.
    :rtype: list({0}(in_datasets)), list({0}(out_datasets))
    """


def mData_notes():
    """ Get a list of meta_data objects associated with the {0}_datasets.

    :returns: All MetaData objects associated with {0} data objects.
    :rtype: list(MetaData)
    """



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: chunking_tests
   :platform: Unix
   :synopsis: checking the chunking for a variety of pattern transforms

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import unittest
from savu.test import test_utils as tu
import numpy as np

from savu.data.chunking import Chunking
from savu.data.experiment_collection import Experiment


class ChunkingTests(unittest.TestCase):
    def create_chunking_instance(self, current_list, nnext_list, nProcs):
        current = self.create_pattern('a', current_list)
        nnext = self.create_pattern('b', nnext_list)
        options = tu.set_experiment('tomoRaw')
        options['processes'] = list(range(nProcs))
        # set a dummy process list
        options['process_file'] = \
            tu.get_test_process_path('loaders/basic_tomo_process.nxs')
        exp = Experiment(options)
        test_dict = {'current': current, 'next': nnext}
        chunking = Chunking(exp, test_dict)
        return chunking

    def create_pattern(self, name, pattern_list):
        pattern = {name: {'max_frames_transfer': pattern_list[0],
                          'slice_dims': pattern_list[1],
                          'core_dims': pattern_list[2]}}
        return pattern

    def get_nChunks(self, shape, chunks):
        prod = 1
        for i in range(len(shape)):
            prod *= np.ceil(shape[i]/float(chunks[i]))
        return prod

    def amend_chunks(self, chunks):
        """ If any temporary amendments are applied to the final chunking
        values in the framework, then remove these here. """
        return chunks

    def test_chunks_3D_1(self):
        current = [1, (0,), (1, 2)]
        nnext = [1, (0,), (1, 2)]
        shape = (5000, 500, 500)
        nProcs = 1
        chunking = self.create_chunking_instance(current, nnext, nProcs)
        chunks = chunking._calculate_chunking(shape, np.float32)
        self.assertEqual(self.amend_chunks(chunks), (1, 500, 500))

        nProcs = 2
        chunking = self.create_chunking_instance(current, nnext, nProcs)
        chunks = chunking._calculate_chunking(shape, np.float32)
        self.assertEqual(self.amend_chunks(chunks), (1, 500, 500))

        shape = (5000, 5000, 5000)
        chunking = self.create_chunking_instance(current, nnext, nProcs)
        chunks = chunking._calculate_chunking(shape, np.float32)
        self.assertEqual(self.amend_chunks(chunks), (1, 625, 313))

        shape = (5000, 5000, 5000)
        chunking = self.create_chunking_instance(current, nnext, nProcs)
        chunks = chunking._calculate_chunking(shape, np.float32)
        self.assertEqual(self.amend_chunks(chunks), (1, 625, 313))

        shape = (1, 800, 500)
        chunking = self.create_chunking_instance(current, nnext, nProcs)
        chunks = chunking._calculate_chunking(shape, np.float32)
        self.assertEqual(self.amend_chunks(chunks), (1, 400, 500))

    def test_chunks_3D_2(self):
        current = [1, (0,), (1, 2)]
        nnext = [1, (1,), (0, 2)]
        shape = (50, 300, 100)
        nProcs = 1
        chunking = self.create_chunking_instance(current, nnext, nProcs)
        chunks = chunking._calculate_chunking(shape, np.float32)
        self.assertEqual(self.amend_chunks(chunks), (50, 50, 100))

        current = [8, (0,), (1, 2)]
        nnext = [4, (1,), (0, 2)]
        chunking = self.create_chunking_instance(current, nnext, nProcs)
        chunks = chunking._calculate_chunking(shape, np.float32)
        self.assertEqual(self.amend_chunks(chunks), (48, 52, 100))

        nProcs = 10
        chunking = self.create_chunking_instance(current, nnext, nProcs)
        chunks = chunking._calculate_chunking(shape, np.float32)
        self.assertEqual(self.amend_chunks(chunks), (8, 28, 100))

    def test_chunks_4D_1(self):
        current = [1, (0, 1), (2, 3)]
        nnext = [1, (0, 1), (2, 3)]
        shape = (800, 700, 600, 500)
        nProcs = 1
        chunking = self.create_chunking_instance(current, nnext, nProcs)
        chunks = chunking._calculate_chunking(shape, np.float32)
        self.assertEqual(self.amend_chunks(chunks), (1, 1, 300, 500))

        current = [1, (0, 1), (2, 3)]
        nnext = [1, (2, 3), (0, 1)]
        chunking = self.create_chunking_instance(current, nnext, nProcs)
        chunks = chunking._calculate_chunking(shape, np.float32)
        self.assertEqual(self.amend_chunks(chunks), (23, 22, 22, 22))

        current = [1, (0,), (1, 2, 3)]
        nnext = [1, (0,), (1, 2, 3)]
        chunking = self.create_chunking_instance(current, nnext, nProcs)
        chunks = chunking._calculate_chunking(shape, np.float32)
        self.assertEqual(self.amend_chunks(chunks), (1, 44, 75, 63))

        current = [4, (0,), (1, 2, 3)]
        nnext = [8, (1, 2), (0, 3)]
        chunking = self.create_chunking_instance(current, nnext, nProcs)
        chunks = chunking._calculate_chunking(shape, np.float32)
        self.assertEqual(self.amend_chunks(chunks), (8, 8, 7, 500))

        nProcs = 200
        current = [4, (0,), (1, 2, 3)]
        nnext = [8, (1, 2), (0, 3)]
        chunking = self.create_chunking_instance(current, nnext, nProcs)
        chunks = chunking._calculate_chunking(shape, np.float32)
        self.assertEqual(self.amend_chunks(chunks), (4, 8, 8, 500))


if __name__ == "__main__":
    unittest.main()
